
import rclpy
from ravestate import registry
from ravestate.constraint import s
from ravestate.state import state
from ravestate.receptor import receptor
from ravestate.property import PropertyBase
from std_msgs.msg import String


rclpy.init()
node = rclpy.create_node("vision_node")

@state(cond=s(":startup"))
def facerec_run(ctx):

    @receptor(ctx_wrap=ctx, write="facerec:face")
    def face_recognition_callback(ctx, msg):
        ctx["facerec:face"] = msg

    node.create_subscription(String, "/roboy/vision/recognized_faces", face_recognition_callback)
    rclpy.spin(node)


@state(cond=s(":shutdown"))
def facerec_shutdown():
    node.destroy_node()
    rclpy.shutdown()


registry.register(
    name="facerec",
    props=PropertyBase(name="face", default_value=""),
    states=(facerec_run, facerec_shutdown))


# Ravestate class which encapsulates a graph of signal parent/offspring instances

from typing import Set, Dict, Optional, List
from ravestate.iactivation import IActivation, ISpike
from threading import Lock
from collections import defaultdict

from reggol import get_logger
logger = get_logger(__name__)


class CausalGroup:
    """
    Class which represents a causal group graph of spike parent/offspring
     spikes (a "superspike"). These must synchronize wrt/ their (un)written
     properties and state activation candidates, such that they don't cause output races.

    Note: Always use a `with ...` construct to interact with a causal group.
     Otherwise, undefined behavior may occur due to race conditions.
    """

    # Lock for the whole causal group
    _lock: Lock
    # Remember the locked lock, since the lock member might be re-targeted in merge()
    _locked_lock: Optional[Lock]

    # Set of property names for which no writing state has been
    #  activated yet within this causal group.
    #
    # This is different from _ref_index:
    #
    #  -> _unwritten_props might lack properties that are only
    #   *temporarily* unavailable, because a writing state is still running,
    #   and might still resign, which would make that state's write prop's
    #   available again.
    #
    #  -> In this case it is useful to still see the other
    #   state activation candidates for those props in _ref_index.
    #
    # Also, _ref_index needs to be as
    #  sparse as possible to optimise performance.
    #
    _available_resources: Set[str]

    # Refcount per state activations per spike per property name.
    #  Refcount, because one activation may hold multiple references to one spike for multiple
    #  differently timed constraints. We essentially keep a {(propname, spike, activation, refcount)}
    #  index structure, that will be necessary for the following use-cases:
    #
    #  * Determining whether an activation is the most specific for a certain property set
    #    -> Gather state activations per spike (i) per property (p), sort by specificity
    #    -> O(i * log p)
    #  * Determining whether a spike is stale (no more activation references)
    #    -> For every property name (p), check whether the spike (i) is registered
    #    -> O(p * log i)
    #
    #  Container operations:
    #
    #  * Inserting a spike (i) / state activation (a)
    #    -> O(log p + log i + log a)
    #  * Removing a state activation (a) for a spike (i) (for each of the state's referenced properties (p))
    #    -> O(log p + log i + log a)
    #  * Wiping a spike (i)
    #    -> O(p * log i)
    #
    _ref_index: Dict[
        str,  # Resource name
        Dict[
            ISpike,
            Dict[
                IActivation,
                int
            ]
        ]
    ]

    # Names of member spikes for __repr__, added by Spike ctor
    signal_names: List[str]

    # Object to count the number of CausalGroups merged in this instance
    merges: List[int]

    def __init__(self, resources: Set[str]):
        """
        Create a new causal group, with a set of unwritten props.
        """
        self.signal_names = []
        self.merges = [1]
        self._lock = Lock()
        self._locked_lock = None
        self._available_resources = resources.copy()
        self._ref_index = {
            # TODO: Create refcount prop entries on demand
            prop: defaultdict(lambda: defaultdict(int))
            for prop in resources
        }

    def __del__(self):
        with self._lock:
            self.merges[0] -= 1
        logger.debug(f"Deleted {self}")

    def __enter__(self) -> 'CausalGroup':
        # Remember current lock, since it might change,
        # if the lock object is switched in merge()
        lock = self._lock
        lock.__enter__()
        if lock != self._lock:
            lock.__exit__(None, None, None)
            return self.__enter__()
        # Remember the locked lock, since the lock member might be re-targeted in merge()
        assert not self._locked_lock
        self._locked_lock = self._lock
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> bool:
        assert self._locked_lock
        result = self._locked_lock.__exit__(exc_type, exc_value, traceback)
        self._locked_lock = None
        return result

    def __eq__(self, other) -> bool:
        return isinstance(other, CausalGroup) and other._lock == self._lock

    def __hash__(self):
        return self._lock.__hash__()

    def __repr__(self):
        if len(self.signal_names) < 4:
            spikes = ','.join(self.signal_names)
        else:
            spikes = f"{self.signal_names[0]},...[{len(self.signal_names)-2} more],{self.signal_names[-1]}"
        return f"CausalGroup*{self.merges[0]}@{hex(id(self._lock))[2:]}({spikes})"

    def merge(self, other: 'CausalGroup'):
        """
        Merge this causal group with another. Unwritten props will become
         the set intersection of this group's unwritten props and
         other's unwritten props. consumed() will be called with
         all properties that are consumed by other, but not this.
        Afterwards, other's member objects will be set to this's.
        """
        # Intersect _available_resources
        self._available_resources = self._available_resources & other._available_resources

        # Intersect/merge _ref_index
        allowed_propnames = self._ref_index.keys() & other._ref_index.keys()
        self.consumed(set(self._ref_index.keys()) - allowed_propnames)
        other.consumed(set(other._ref_index.keys()) - allowed_propnames)
        for prop, spikes in other._ref_index.items():
            for spike, act_refcount in spikes.items():
                for act, refcount in act_refcount.items():
                    if refcount > 0:
                        self._ref_index[prop][spike][act] += refcount

        # Merge signal names/merge count
        self.signal_names += other.signal_names
        self.merges[0] += other.merges[0]

        # Retarget other's members
        other.signal_names = self.signal_names
        other.merges = self.merges
        other._lock = self._lock
        other._available_resources = self._available_resources
        other._ref_index = self._ref_index

    def acquired(self, spike: 'ISpike', acquired_by: IActivation) -> bool:
        """
        Called by Activation to notify the causal group, that
         it is being referenced by an activation constraint for a certain member spike.

        * `spike`: State activation instance, which is now being
         referenced by the specified causal group.

        * `acquired_by`: State activation instance,
         which is interested in this property.

        **Returns:** Returns True if all of the acquiring's write-props are
         free, and the group now refs. the activation, False otherwise.
        """
        # Make sure that all properties are actually still writable
        for prop in acquired_by.resources():
            if prop not in self._ref_index:
                logger.error(f"{prop} is unavailable, but {acquired_by} wants it from {self}!")
                return False
        for prop in acquired_by.resources():
            self._ref_index[prop][spike][acquired_by] += 1
        return True

    def rejected(self, spike: 'ISpike', rejected_by: IActivation, reason: int) -> None:
        """
        Called by a state activation, to notify the group that a member spike
         is no longer being referenced for the given state's write props.
        This may be because ... <br>
        ... the state activation auto-eliminated. (reason=0) <br>
        ... the spike got too old. (reason=1) <br>
        ... the activation is happening and dereferencing it's spikes. (reason=2)

        * `spike`: The member spike whose ref-set should be reduced.

        * `rejected_by`: State activation instance, which is no longer
         interested in this property.

        * `reason`: See about.
        """
        for prop in rejected_by.resources():
            if prop in self._ref_index and \
               spike in self._ref_index[prop] and \
               rejected_by in self._ref_index[prop][spike]:
                remaining_refcount = self._ref_index[prop][spike][rejected_by]
                if remaining_refcount > 0:
                    self._ref_index[prop][spike][rejected_by] -= 1
                    if remaining_refcount > 1:
                        # do not fall through to ref deletion
                        continue
                else:
                    logger.error(f"Attempt to deref group for unref'd activation {rejected_by.name}")

                del self._ref_index[prop][spike][rejected_by]
                if len(self._ref_index[prop][spike]) == 0:
                    del self._ref_index[prop][spike]
        if reason == 1:
            logger.debug(f"{self}.rejected({spike} by {rejected_by} due to age)")

    def consent(self, ready_suitor: IActivation) -> bool:
        """
        Called by constraint, to inquire whether this causal group would happily
         be consumed for the given state activation's properties.
        This will be called periodically on the group by state activations
         that are ready to go. Therefore, a False return value from this
         function is never a final judgement (more like a "maybe later").

        * `ready_suitor`: The state activation which would like to consume
         this instance for it's write props.

        **Returns:** True if this instance agrees to proceeding with the given consumer
         for the consumer's write props, False otherwise.
        """
        specificity = ready_suitor.specificity()
        higher_specificity_acts = set()
        highest_higher_specificity = .0
        highest_higher_specificity_act = None
        for prop in ready_suitor.resources():
            if prop in self._available_resources:
                for spike in self._ref_index[prop]:
                    for candidate in self._ref_index[prop][spike]:
                        if candidate not in higher_specificity_acts:
                            cand_specificity = candidate.specificity()
                            if cand_specificity > specificity:
                                higher_specificity_acts.add(candidate)
                                if cand_specificity > highest_higher_specificity:
                                    highest_higher_specificity = cand_specificity
                                    highest_higher_specificity_act = candidate
            else:
                # Easy exit condition: prop not free for writing
                logger.debug(f"{self}.consent({ready_suitor})->N: {prop} unavailable.")
                return False

        if higher_specificity_acts:
            highest_higher_specificity_act.pressure()
            logger.debug(
                f"{self}.consent({ready_suitor})->N: "
                f"{str(specificity)[:4]} < {str(highest_higher_specificity)[:4]} "
                f"({highest_higher_specificity_act} + {len(higher_specificity_acts)-1} others)")
            return False

        logger.debug(f"{self}.consent({ready_suitor})->Y")
        return True

    def activated(self, act: IActivation):
        """
        Called by activation which previously received a go-ahead
         from consent(), when it is truly proceeding with
         running (after it got the go-ahead from all it's depended=on
         causal groups).

        * `act`: The activation that is now running.
        """
        # Mark the consented w-props as unavailable
        self._available_resources -= act.resources()

    def resigned(self, act: IActivation) -> None:
        """
        Called by activation, to let the causal group know that it failed,
         and a less specific activation may now be considered for
         the resigned state's write props.

        * `act`: The activation that is unexpectedly not consuming it's resources,
         because it's state resigned/failed.
        """
        # Mark the act's w-props as available again
        self._available_resources |= act.resources()

    def consumed(self, resources: Set[str]) -> None:
        """
        Called by activation to notify the group, that it has been
         consumed for the given set of properties.

        * `resources`: The properties which have been consumed.
        """
        if not resources:
            return
        for prop in resources:
            # Notify all concerned activations, that the
            # spikes they are referencing are no longer available
            for sig in self._ref_index[prop]:
                for act in self._ref_index[prop][sig]:
                    act.dereference(spike=sig, reacquire=True)
            # Remove the consumed prop from the index
            del self._ref_index[prop]
        logger.debug(f"{self}.consumed({resources})")

    def wiped(self, spike: 'ISpike') -> None:
        """
        Called by a spike, to notify the causal group that
         the instance was wiped and should no longer be remembered.

        * `spike`: The instance that should be henceforth forgotten.
        """
        for prop in self._ref_index:
            if spike in self._ref_index[prop]:
                for act in self._ref_index[prop][spike]:
                    act.dereference(spike=spike, reacquire=True)
                del self._ref_index[prop][spike]

    def stale(self, spike: 'ISpike') -> bool:
        """
        Determine, whether a spike is stale (has no
        remaining interested activations and no children).

        **Returns:** True, if no activations reference the given
         spike for any unwritten property. False otherwise.
        """
        for prop in self._ref_index:
            if spike in self._ref_index[prop]:
                if len(self._ref_index[prop][spike]) > 0:
                    return False
                else:
                    # Do some cleanup
                    del self._ref_index[prop][spike]
        result = not spike.has_offspring()
        return result



# Ravestate property classes

from threading import Lock
from typing import Dict, List, Generator
from ravestate.constraint import s, Signal

from reggol import get_logger
logger = get_logger(__name__)


class PropertyBase:
    """
    Base class for context properties. Controls read/write/push/pop/delete permissions,
    property name basic impls. for the property value, parent/child mechanism.
    """

    def __init__(
            self, *,
            name="",
            allow_read=True,
            allow_write=True,
            allow_push=True,
            allow_pop=True,
            default_value=None,
            always_signal_changed=False):

        self.name = name
        self.allow_read = allow_read
        self.allow_write = allow_write
        self.allow_push = allow_push
        self.allow_pop = allow_pop
        self.value = default_value
        self.children: Dict[str, PropertyBase] = dict()
        self._lock = Lock()
        self.parent_path: str = ""
        self.always_signal_changed = always_signal_changed

    def fullname(self):
        return f'{self.parent_path}:{self.name}'

    def set_parent_path(self, path):
        """
        Set the ancestors (including modulename) for a property

        * `path`: ancestry in the form of modulename:parent_prop_name (or simply modulename)
        """
        if not self.parent_path:
            self.parent_path = path
        else:
            logger.error(f'Tried to override parent_path of {self.fullname()}')

    def gather_children(self) -> List['PropertyBase']:
        """
        Collect this property, and all of it's children.
        """
        result = [self]
        for child in self.children.values():
            result += child.gather_children()
        return result

    def lock(self):
        self._lock.acquire()

    def unlock(self):
        self._lock.release()

    def read(self):
        """
        Read the current property value
        """
        if not self.allow_read:
            logger.error(f"Unauthorized read access in property {self.fullname()}!")
            return None
        return self.value

    def write(self, value):
        """
        Write a new value to the property

        * `value`: The new value.

        **Returns:** True if the value has changed and :changed should be signaled, false otherwise.
        """
        if not self.allow_write:
            logger.error(f"Unauthorized write access in property {self.fullname()}!")
            return False
        if self.always_signal_changed or self.value != value:
            self.value = value
            return True
        else:
            return False

    def push(self, child: 'PropertyBase'):
        """
        Add a child to the property

        * `child`: The child object

        **Returns:** True if the child was added successfully, false otherwise.
        """
        if not self.allow_push:
            logger.error(f"Unauthorized push in property {self.fullname()}!")
            return False
        if child.name in self.children:
            logger.error(f"Tried to add already existing child-property {self.fullname()}:{child.name}")
            return False
        child.set_parent_path(self.fullname())
        self.children[child.name] = child
        return True

    def pop(self, child_name: str):
        """
        Remove a child from the property by it's name.

        * `child_name`: Name of the child to be removed.

        **Returns:** True if the pop was successful, False otherwise
        """
        if not self.allow_pop:
            logger.error(f"Unauthorized pop in property {self.fullname()}!")
            return False
        elif child_name in self.children:
            self.children.pop(child_name)
            return True
        else:
            logger.error(f"Tried to remove non-existent child-property {self.fullname()}:{child_name}")
            return False

    def changed_signal(self) -> Signal:
        """
        Signal that is emitted by PropertyWrapper when #write() returns True.
        """
        return s(f"{self.fullname()}:changed")

    def pushed_signal(self) -> Signal:
        """
        Signal that is emitted by PropertyWrapper when #push() returns True.
        """
        return s(f"{self.fullname()}:pushed")

    def popped_signal(self) -> Signal:
        """
        Signal that is emitted by PropertyWrapper when #pop() returns True.
        """
        return s(f"{self.fullname()}:popped")

    def signals(self) -> Generator[Signal, None, None]:
        """
        Yields all signals that may be emitted because of
         this property, given it's write/push/pop permissions.
        """
        if self.allow_write:
            yield self.changed_signal()
        if self.allow_push:
            yield self.pushed_signal()
        if self.allow_pop:
            yield self.popped_signal()



from ravestate_nlp.question_word import QuestionWord
from ravestate_nlp.triple import Triple
from spacy.tokens import Token

SUBJECT_SET = {"nsubj"}
OBJECT_SET = {"dobj", "attr", "advmod"}
PREDICATE_SET = {"ROOT", "conj"}
PREDICATE_AUX_SET = {"acomp", "aux", "xcomp"}
VERB_AUX_SET = {"VERB", "ADJ"}
RECURSION_BLACKLIST = {"cc", "conj"}


def extract_triples(doc):
    """
    triple: subject, predicate, object
    finding the triple of a sentence:
    this is done by using the dependency tree of the sentence
    a dependency tree shows the relation between the words of a sentence
    the finite verb (predicate) is the structural center of the sentence and therefor of the tree
    so starting with the predicate the algorithm searches through the dependency tree to find subject and object
    -> triple_search()

    examples:
    - How are you?
     tree depth = 1     predicate: "are"
                            -> relation to "how": adverbial modifier (advmod) => object of this sentence
                            -> relation to "you": nominal subject (nsubj) => subject of this sentence

    - How old are you?
    tree depth = 2      predicate: "are"
                            -> relation to "old": adjectival complement (acomp) => auxiliary predicate
                                -> relation to "how":  adverbial modifier (advmod) => object of this sentence
                            -> relation to "you": nominal subject (nsubj) => subject of this sentence

    - Who is your father?
    tree depth = 2      predicate: "is"
                            -> relation to "who": nominal subject (nsubj) => subject of this sentence
                            -> relation to "father": attribute (attr) => object of this sentence
                                    ->relation to "your": possession modifier (poss)
    """
    triples = []
    for token in doc:
        if token.dep_ in PREDICATE_SET:
            triple = triple_search(Triple(predicate=token), token)
            triple.ensure_notnull(doc._.empty_token)
            triples.append(triple)
    return triples


def triple_search(triple: Triple, token: Token):
    """
    Recursive search through the dependency tree
    looks for triple values in each of the children and calls itself with the children nodes
    """
    question_word = None
    for word in token.children:
        if word.text.lower() in QuestionWord.question_words:
            question_word = QuestionWord(word)
            word = QuestionWord(word)
            if not triple.get_object():
                triple.set_object(question_word)
        elif word.dep_ in OBJECT_SET:
            triple.set_object(word)
        if word.dep_ in SUBJECT_SET:
            triple.set_subject(word)
        if word.dep_ in PREDICATE_AUX_SET and word.pos_ in VERB_AUX_SET:
            triple.set_predicate_aux(word)
        if isinstance(word, Token) and word.dep_ not in RECURSION_BLACKLIST:
            triple = triple_search(triple, word)
    if not triple.get_subject() and question_word:
        triple.set_subject(question_word)
    return triple



from ravestate import registry
from ravestate_telegramio import telegram_bot

import ravestate_rawio

registry.register(
    name="telegramio",
    states=(telegram_bot.telegram_run, telegram_bot.telegram_output),
    config={
        telegram_bot.TOKEN_CONFIG_KEY: ""
    })



# Ravestate receptor state decorator

from ravestate.state import State
from ravestate.wrappers import ContextWrapper
from ravestate.activation import Activation
from typing import Union, Set, Tuple


def receptor(*, ctx_wrap: ContextWrapper, write: Union[str, Tuple[str]]):
    """
    A receptor is a special state which can be invoked from outside,
     to push values into the context.

    * `ctx_wrap`: A context wrapper as is always given
     into the state functions as their first argument.

    * `write`: The property, or tuple of properties, which are going to be written.
    """

    def receptor_decorator(action):
        nonlocal ctx_wrap, write
        ctx = ctx_wrap.ctx
        receptor_state = State(
            write=write,
            read=(),
            signal_name=None,
            cond=None,
            action=action,
            is_receptor=True)

        def receptor_function(*args, **kwargs):
            nonlocal receptor_state, ctx
            Activation(st=receptor_state, ctx=ctx).run(*args, **kwargs)

        return receptor_function

    return receptor_decorator



from ravestate.state import state
from ravestate import registry
from ravestate.constraint import s
from ravestate.wrappers import ContextWrapper

import ravestate_verbaliser
import ravestate_phrases_basic_en
import ravestate_interloc


@state(cond=s("interloc:all:pushed") & s("rawio:in:changed"), write="verbaliser:intent")
def react_to_pushed_interloc(ctx: ContextWrapper):
    ctx["verbaliser:intent"] = "greeting"


@state(cond=s("interloc:all:popped") & s("rawio:in:changed"), write="verbaliser:intent")
def react_to_popped_interloc(ctx: ContextWrapper):
    ctx["verbaliser:intent"] = "farewells"


registry.register(name="hi_goodbye", states=(react_to_pushed_interloc, react_to_popped_interloc))



from spacy.tokens import Token


class QuestionWord:
    FORM = 'FORM'
    CHOICE = 'CHOICE'
    OBJECT = 'OBJECT'
    REASON = 'REASON'
    TIME = 'TIME'
    PERSON = 'PERSON'
    PLACE = 'PLACE'

    question_pos = 'QUESTION'
    question_words = {
        'where': PLACE,
        'who': PERSON,
        'when': TIME,
        'why': REASON,
        'what': OBJECT,
        'which': CHOICE,
        'how': FORM
    }

    def __init__(self, token: Token):
        self.text = self.question_words[token.text.lower()]
        self.lemma_ = self.question_words[token.text.lower()]
        self.pos_ = self.question_pos
        self.dep_ = token.dep_
        self.is_space = False



from spacy.tokens import Token
from typing import Optional, Set, Union, Tuple
from ravestate_nlp.question_word import QuestionWord


class Triple:

    _subject: Union[Token, QuestionWord]
    _predicate: Token
    _predicate_aux: Token
    _object: Union[Token, QuestionWord]

    def __init__(self, subject: Token = None, predicate: Token = None, predicate_aux: Token=None, object: Token = None):
        self.set_subject(subject)
        self.set_predicate(predicate)
        self.set_predicate_aux(predicate_aux)
        self.set_object(object)

    def set_subject(self, subject: Union[Token, QuestionWord]):
        self._subject = subject

    def set_predicate(self, predicate: Token):
        self._predicate = predicate
    
    def set_predicate_aux(self, predicate_aux: Token):
        self._predicate_aux = predicate_aux

    def set_object(self, object: Union[Token, QuestionWord]):
        self._object = object

    def get_subject(self) -> Union[Token, QuestionWord]:
        return self._subject

    def get_predicate(self) -> Token:
        return self._predicate

    def get_predicate_aux(self) -> Token:
        return self._predicate_aux

    def get_object(self) -> Union[Token, QuestionWord]:
        return self._object

    def to_tuple(self) -> Tuple[str, str, str]:
        sub = ""
        pre = ""
        obj = ""
        if self._subject and not self._subject.is_space:
            sub = self._subject.text
        if self._predicate and not self._predicate.is_space:
            pre = self._predicate.lemma_
        if self._predicate_aux and not self._predicate_aux.is_space:
            pre += " " + self._predicate_aux.lemma_
        if self._object and not self._object.is_space:
            obj = self._object.text
        return sub, pre, obj

    def match_either_lemma(self, pred: Optional[Set[str]] = None, subj: Optional[Set[str]] = None, obj: Optional[Set[str]] = None):
        if pred and (self._predicate.lemma_ in pred or self._predicate_aux.lemma_ in pred):
            return True
        if subj and self._subject.lemma_ in subj:
            return True
        if obj and self._object.lemma_ in obj:
            return True
        return False

    def is_question(self, question_word: Optional[str] = None):
        if question_word:
            return self.match_either_lemma(obj={question_word}, subj={question_word})
        else:
            return isinstance(self._subject, QuestionWord) or isinstance(self._object, QuestionWord)

    def ensure_notnull(self, empty_token):
        # do not allow empty entries in triple
        if not self._subject:
            self._subject = empty_token
        if not self._predicate:
            self._predicate = empty_token
        if not self._predicate_aux:
            self._predicate_aux = empty_token
        if not self._object:
            self._object = empty_token

    def __eq__(self, other) -> bool:
        if isinstance(other, Triple):
            return str(self).lower() == str(other).lower()
        elif isinstance(other, tuple) or isinstance(other, list):
            if len(other) == 3:
                return \
                    str(other[0]).lower() == self._subject.text.lower() and \
                    str(other[1]).lower() == self._predicate.lemma_.lower() and \
                    str(other[2]).lower() == self._object.lemma_.lower()
        return False

    def __str__(self):
        return ":".join(self.to_tuple())

    def __repr__(self):
        return f"<Triple object {str(self)}>"




from ravestate import registry
from ravestate.property import PropertyBase
from ravestate.state import state
from ravestate_nlp.question_word import QuestionWord
from ravestate_nlp.triple import Triple
from ravestate_nlp.extract_triples import extract_triples
from ravestate.state import Emit
import spacy


from reggol import get_logger
logger = get_logger(__name__)


def init_model():
    # TODO: Create nlp instance in :startup state, save in context instead of global var
    global nlp, empty_token
    nlp = spacy.load('en_core_web_sm')
    empty_token = nlp(u' ')[0]

    # TODO: Make agent id configurable, rename nlp:contains-roboy to nlp:agent-mentioned
    about_roboy = ('you', 'roboy', 'robot', 'roboboy', 'your')
    def roboy_getter(doc) -> bool:
        return any(roboy in doc.text.lower() for roboy in about_roboy)

    from spacy.tokens import Doc
    Doc.set_extension('about_roboy', getter=roboy_getter)
    Doc.set_extension('empty_token', getter=lambda doc: empty_token)
    Doc.set_extension('triples', getter=extract_triples)


@state(read="rawio:in", write=("nlp:tokens", "nlp:postags", "nlp:lemmas", "nlp:tags", "nlp:ner", "nlp:triples", "nlp:roboy","nlp:triples"))
def nlp_preprocess(ctx):
    nlp_doc = nlp(ctx["rawio:in"])
    
    nlp_tokens = tuple(str(token) for token in nlp_doc)
    ctx["nlp:tokens"] = nlp_tokens
    logger.info(f"[NLP:tokens]: {nlp_tokens}")

    nlp_postags = tuple(str(token.pos_) for token in nlp_doc)
    ctx["nlp:postags"] = nlp_postags
    logger.info(f"[NLP:postags]: {nlp_postags}")

    nlp_lemmas = tuple(str(token.lemma_) for token in nlp_doc)
    ctx["nlp:lemmas"] = nlp_lemmas
    logger.info(f"[NLP:lemmas]: {nlp_lemmas}")
    
    nlp_tags = tuple(str(token.tag_) for token in nlp_doc)
    ctx["nlp:tags"] = nlp_tags
    logger.info(f"[NLP:tags]: {nlp_tags}")

    nlp_ner = tuple((str(ents.text), str(ents.label_)) for ents in nlp_doc.ents)
    ctx["nlp:ner"] = nlp_ner
    logger.info(f"[NLP:ner]: {nlp_ner}")

    nlp_triples = nlp_doc._.triples
    ctx["nlp:triples"] = nlp_triples
    logger.info(f"[NLP:triples]: {nlp_triples}")

    nlp_roboy = nlp_doc._.about_roboy
    ctx["nlp:roboy"] = nlp_roboy
    logger.info(f"[NLP:roboy]: {nlp_roboy}")


@state(signal_name="contains-roboy", read="nlp:roboy")
def nlp_contains_roboy_signal(ctx):
    if ctx["nlp:roboy"]:
        return Emit()
    return False


@state(signal_name="is-question", read="nlp:triples")
def nlp_is_question_signal(ctx):
    if ctx["nlp:triples"][0].is_question():
        return Emit()
    return False


init_model()
registry.register(
    name="nlp",
    states=(
        nlp_preprocess,
        nlp_contains_roboy_signal,
        nlp_is_question_signal
    ),
    props=(
        PropertyBase(name="tokens", default_value="", always_signal_changed=True, allow_pop=False, allow_push=False),
        PropertyBase(name="postags", default_value="", always_signal_changed=True, allow_pop=False, allow_push=False),
        PropertyBase(name="lemmas", default_value="", always_signal_changed=True, allow_pop=False, allow_push=False),
        PropertyBase(name="tags", default_value="", always_signal_changed=True, allow_pop=False, allow_push=False),
        PropertyBase(name="ner", default_value="", always_signal_changed=True, allow_pop=False, allow_push=False),
        PropertyBase(name="triples", default_value="", always_signal_changed=True, allow_pop=False, allow_push=False),
        PropertyBase(name="roboy", default_value="", always_signal_changed=True, allow_pop=False, allow_push=False)
    )
)



# Interface of state activation towards spike
# Spike interface towards state activation

from typing import Set, Optional, Generator


class ISpike:
    """
    Base interface class for spikes.
    """

    def name(self) -> str:
        """
        Returns the name of this spike's signal.
        """
        pass

    def wipe(self, already_wiped_in_causal_group: bool=False) -> None:
        """
        Called either in Context run loop when the spike is found to be stale
         (with wiped_in_causal_group=True), or in Context.wipe(signal_inst),
         or by parent (recursively).
        After this function is called, the spike should be cleaned up by GC.

        * `already_wiped_in_causal_group`: Boolean which indicates, whether wiped(signal_inst)
         must still be called on the group to make sure sure that no dangling references
         to the spike are maintained by any state activations.
        """
        pass

    def has_offspring(self):
        """
        Called by CausalGroup.stale(spike).

        **Returns:** True if the spike has active offspring, false otherwise.
        """
        pass

    def offspring(self) -> Generator['Spike', None, None]:
        """
        Recursively yields this spike's offspring and it's children's offspring.

        **Returns:** All of this spike's offspring spikes.
        """
        yield None


class IActivation:
    """
    Base interface class for state activations.
    """

    name: str

    def resources(self) -> Set[str]:
        """
        Return's the set of the activation's write-access property names.
        """
        pass

    def specificity(self) -> float:
        """
        Returns the lowest specificity among the specificity values of the
         activation's conjunct contraints. The specificity for a single conjunction
         is calculated as the sum of it's component signal's specificities,
         which in turn is calculated as one over the signal's subscriber count.
        """
        pass

    def dereference(self, *, spike: Optional[ISpike]=None, reacquire: bool=False, reject: bool=False) -> None:
        """
        Notify the activation, that a single or all spike(s) are not available
         anymore, and should therefore not be referenced anymore by the activation.
        This is called by ... <br>
         ... context when a state is deleted. <br>
         ... causal group, when a referenced signal was consumed for a required property. <br>
         ... causal group, when a referenced signal was wiped. <br>
         ... this activation (with reacquire=True), if it gives in to activation pressure.

        * `spike`: The spike that should be forgotten by the activation, or
         none, if all referenced spikes should be forgotten.

        * `reacquire`: Flag which tells the function, whether for every rejected
         spike, the activation should hook into context for reacquisition
         of a replacement spike.

        * `reject`: Flag which controls, whether de-referenced spikes
         should be explicitely rejected through their causal groups.
        """
        pass

    def pressure(self):
        """
        Called by CausalGroup, to pressure the activation to
         make a decision on whether it is going to retain a reference
         to the given spike, given that there is a lower-
         specificity activation which is ready to run.
        """
        pass

    def secs_to_ticks(self, seconds: float) -> int:
        """
        Convert seconds to an equivalent integer number of ticks,
         given this activation's context's tick rate.

        * `seconds`: Seconds to convert to ticks.

        **Returns:** An integer tick count.
        """
        pass



# Ravestate argument parser

from argparse import ArgumentParser
from argparse import RawDescriptionHelpFormatter
from argparse import Action
from typing import List, Tuple, Any

from reggol import get_logger, help_string
logger = get_logger(__name__)


def handle_args(*args) -> Tuple[List[str], List[Tuple[str, str, Any]], List[str]]:
    """
    Runs an argument parser for the given args. Returns modules-to-load,
    config value-overrides and config file-pathes.
     Note: If the arguments are ill-formatted, or the -h argument is passed,
    help will be printed to the console and the program will abort.

    * `args`: Argument list which will be fed into argparse.parse_args.

    **Returns:** A Tuple with three items:
    1.) A list of module names which should be imported.
    2.) A list of tuples, where each tuple is a module name, a config key name, and a value.
    3.) A list of yaml file paths.
    """
    modules_to_import: List[str] = []
    config_value_overrides: List[Tuple[str, str, Any]] = []
    yaml_file_paths: List[str] = []

    argdef = ArgumentParser(
        description="Run a reactive state machine! It's weird, you'll like it.",
        formatter_class=RawDescriptionHelpFormatter,
        epilog=f"""
{help_string()}

usage:
  > rasta ravestate_facerec ravestate_hello_world
    Import two python modules and run a context.

  > rasta \\
      -d core import ravestate_facerec ravestate_hello_world \\
      -f generic.yml \\
      -f user.yml
    Import two python modules and run a context,
    with additional config entries loaded from generic.yml
    and user.yml, where entries in user.yml overload entries
    in generic.yml.
"""
    )

    class ModulesAction(Action):
        def __call__(self, parser, args, values, option_string=None):
            nonlocal modules_to_import
            if isinstance(values, str):
                values = [values]
            modules_to_import += values

    class OverrideAction(Action):
        def __call__(self, parser, args, values, option_string=None):
            nonlocal config_value_overrides
            if isinstance(values, str):
                values = [values]
            if len(values) < 3:
                logger.error(f"Not enough values for -d argument: expecting 3, got {len(values)}!")
                return
            if len(values) > 3:
                values = (values[0], values[1], tuple(values[2:]))
            else:
                values = tuple(values)
            config_value_overrides += [values]

    class ConfigFileAction(Action):
        def __call__(self, parser, args, values, option_string=None):
            nonlocal yaml_file_paths
            if isinstance(values, str):
                values = [values]
            yaml_file_paths += values

    argdef.add_argument(
        "modules", nargs="*", type=str, metavar="python-module-name",
        action=ModulesAction,
        help="""
        List ravestate python modules which should be loaded.
        """)

    argdef.add_argument(
        "-d", "--define",
        nargs="+", type=str, metavar=("module key", "values"),
        action=OverrideAction,
        help="""
        Set the value(s) for a certain config entry. If multiple values are given,
        they will be converted to a list. An error will be raised if the given
        module does not exist, or hasn't declared the specified config entry,
        or the given value type (list/scalar) doesn't match the declared.
        """)

    argdef.add_argument(
        "-f", "--config",
        nargs=1, type=str, metavar="yaml-path",
        action=ConfigFileAction,
        help="""
        Add a yaml file path, from which config entries for a certain
        module should be read. The file should contain a `module` key
        which specifies the target module name, and a `config` key which
        is a dictionary of config values.
         Note, that config entries specified in later yaml files override
        entries from earlier files.
        """)
    argdef.add_help = True
    argdef.parse_args(args)

    return modules_to_import, config_value_overrides, yaml_file_paths



# Ravestate class which encapsulates a single spike

from typing import Set, Generator
from ravestate.iactivation import ISpike
from ravestate.causal import CausalGroup

from reggol import get_logger
logger = get_logger(__name__)


class Spike(ISpike):
    """
    This class encapsulates a single spike, to track ... <br>
    ... it's consumption for different output properties (through #CausalGroup). <br>
    ... it's offspring instances (causal group -> spikes caused by this spike)
    """

    # Age of the spike in ticks
    _age: int

    # Name of the spike's signal
    _name: str

    # This spike's causal group. The causal group
    #  is shared by an spike's family (children/parents),
    #  and dictates which properties are still free to be written.
    _causal_group: CausalGroup

    # Offspring signals which were (partially) caused by this spike,
    #  with which consumed props are synced, and which are wiped if this
    #  spike is wiped.
    _offspring: Set['ISpike']

    # Parent instances, which are notified when this spike is wiped
    _parents: Set['Spike']

    def __init__(self, *, sig: str, parents: Set['Spike']=None, consumable_resources: Set[str]=None):
        """
        Construct a spike from a signal name and a list of causing parent signals.

        * `sig`: Name of the signal which is represented by this spike

        * `parents`: The parent spikes which were involved in causing this spike,
         and with which unwritten properties will be synchronized: This spike's
         causal group will be inferred from the parents, and the parent's causal
         groups will be merged if they are different.

        * `consumable_resources`: The set of property names from context,
         which are available for consumption.
        """
        if parents is None:
            parents = set()
        if consumable_resources is None:
            consumable_resources = set()
        self._name = sig
        self._age = 0
        self._offspring = set()
        self._parents = parents.copy() if parents else set()
        self._causal_group = next(iter(parents)).causal_group() if parents else CausalGroup(consumable_resources)
        self._suitors_per_property = {prop: set() for prop in consumable_resources}
        for parent in parents:
            parent.adopt(self)
        with self._causal_group as cg:
            cg.signal_names.append(sig)

    def __del__(self):
        logger.debug(f"Deleted {self}")

    def __repr__(self):
        return f"Spike({self._name}, age={self._age})"

    def name(self) -> str:
        """
        Returns the name of this spike's signal.
        """
        return self._name

    def causal_group(self) -> CausalGroup:
        """
        Get this spike's causal group.

        **Returns:** This instances causal group. Should never be None.
        """
        return self._causal_group

    def adopt(self, child: 'Spike') -> None:
        """
        Called in spike constructor, for instances which claim to be
         caused by this spike.

        * `child`: The child to add to this spike's causal group.
        """
        self._offspring.add(child)
        if self.causal_group() != child.causal_group():
            with self.causal_group() as causal_parent, child.causal_group() as causal_child:
                causal_parent.merge(causal_child)

    def wiped(self, child: 'ISpike') -> None:
        """
        Called by an offspring signal, to notify the spike
         that it was wiped, and should therefore be removed from the children set.

        * `child`: The child to be forgotten.
        """
        if child not in self._offspring:
            logger.warning(f"Offspring {child.name()} requested to be removed from {self.name()}, but it's unfamiliar!")
            return
        self._offspring.remove(child)

    def wipe(self, already_wiped_in_causal_group: bool=False) -> None:
        """
        Called either in Context run loop when the spike is found to be stale
         (with wiped_in_causal_group=True), or in Context.wipe(spike),
         or by parent (recursively).
        After this function is called, the spike should be cleaned up by GC.

        * `already_wiped_in_causal_group`: Boolean which indicates, whether wiped(spike)
         must still be called on the group to make sure sure that no dangling references
         to the spike are maintained by any state activations.
        """
        # Wipe children. Copy set, because self._offspring will be manipulated during iteration.
        offspring = self._offspring.copy()
        for child in offspring:
            child.wipe(already_wiped_in_causal_group)
        # Notify parents of their child's demise
        for parent in self._parents:
            parent.wiped(self)
        self._parents.clear()
        # Wipe from causal group
        if not already_wiped_in_causal_group:
            with self.causal_group() as causal:
                causal.wiped(self)
        # del self._causal_group

    def has_offspring(self):
        """
        Called by CausalGroup.stale(spike).

        **Returns:** True if the spike has active offspring, false otherwise.
        """
        return len(self._offspring) > 0

    def tick(self) -> None:
        """
        Increment this spike's age by 1.
        """
        self._age += 1

    def age(self) -> int:
        """
        Obtain this spike's age (in ticks).
        """
        return self._age

    def offspring(self) -> Generator['Spike', None, None]:
        """
        Recursively yields this spike's offspring and it's children's offspring.

        **Returns:** All of this spike's offspring spikes.
        """
        for child in self._offspring:
            yield child
            yield from child.offspring()




from ravestate.state import state
from ravestate import registry
from ravestate.constraint import s

from roboy_parlai import wildtalk
import ravestate_rawio


@state(cond=s("rawio:in:changed", max_age=-1), read="rawio:in", write="rawio:out")
def wildtalk_state(ctx):
    ctx["rawio:out"] = wildtalk(ctx["rawio:in"])


registry.register(name="wildtalk", states=(wildtalk_state,))



# Ravestate State-related definitions

from typing import Callable, Optional, Any, Tuple, Union

from ravestate.constraint import Conjunct, Disjunct, Signal, s, Constraint
from ravestate.consumable import Consumable

from reggol import get_logger
logger = get_logger(__name__)


class StateActivationResult:
    """
    Base class for return values of state activation functions.
    """
    pass


class Delete(StateActivationResult):
    """
    Return an instance of this class, if the invoked state should be deleted.

    * `resign`: Set to true, if the state being deleted is due to
     it failing to execute, so a resignation is implied.
     This means, that the spikes that were allocated for it's activation
      may be re-used by another state.
    """
    def __init__(self, resign: bool=False):
        self.resign = resign


class Wipe(StateActivationResult):
    """
    Return an instance of this class, if context.wipe(signal) should be called,
     to ensure that there are no more active spikes for the state's signal.
    """
    pass


class Emit(StateActivationResult):
    """
    Return an instance of this class, if the invoked state's signal should be emitted.

    * `wipe`: Set to true, if context.wipe(signal) should be called before emit,
     to ensure that there is only one free spike for the given signal.
    """
    def __init__(self, wipe: bool=False):
        self.wipe = wipe


class Resign(StateActivationResult):
    """
    Return an instance of this class, if the state invocation should be regarded unsuccessful.
     This means, that the state's signal will not be emitted, and the spikes
     that were allocated for it's activation may be re-used by another state.
    """
    pass


class State:

    _signal: Signal
    signal_name: str
    write_props: Tuple
    read_props: Tuple
    constraint: Constraint
    constraint_: Constraint  # Updated by context, to add constraint causes to constraint
    module_name: str

    # Dummy resource which allows CausalGroups to track acquisitions
    #  for states that don't have any write-props.
    consumable: Consumable

    def __init__(self, *,
                 signal_name: Optional[str],
                 write: Union[str, Tuple[str]],
                 read: Union[str, Tuple[str]],
                 cond: Constraint,
                 action,
                 is_receptor: Optional[bool]=False):

        assert(callable(action))
        self.name = action.__name__
        self.consumable = Consumable(f"@{action.__module__}:{action.__name__}")

        # check to recognize states using old signal implementation
        if isinstance(cond, str):
            logger.error(f"Attempt to create state {self.name} which has a string as condition, not a constraint!")
            cond = None

        # catch the insane case
        if not len(read) and not cond and not is_receptor:
            logger.warning(
                f"The state `{self.name}` is not reading any properties, nor waiting for any signals. " +
                "It will never be activated!")

        # convert read/write properties to tuples
        if isinstance(write, str):
            write = (write,)
        if isinstance(read, str):
            read = (read,)

        # listen to default changed-signals if no signals are given.
        # convert triggers to disjunctive normal form.
        if not cond and len(read) > 0:
            cond = Disjunct(*list(Conjunct(s(f"{rprop_name}:changed")) for rprop_name in read))

        self.signal_name = signal_name
        self.write_props = write
        self.read_props = read
        self.constraint = cond
        self.constraint_ = cond
        self.action = action
        self.module_name = ""
        self._signal = None

    def __call__(self, context, *args, **kwargs) -> Optional[StateActivationResult]:
        args = (context,) + args
        return self.action(*args, **kwargs)

    def signal(self) -> Optional[Signal]:
        assert self.module_name
        if not self._signal and self.signal_name:
            self._signal = s(f"{self.module_name}:{self.signal_name}")
        return self._signal


def state(*, signal_name: Optional[str]="", write: tuple=(), read: tuple=(), cond: Constraint=None):
    """
    Decorator to declare a new state, which may emit a certain signal,
    write to a certain set of properties (calling write, push, pop),
    and read from certain properties (calling read).
    """
    def state_decorator(action):
        nonlocal signal_name, write, read, cond
        return State(signal_name=signal_name, write=write, read=read, cond=cond, action=action)
    return state_decorator



from ravestate import registry
from ravestate.state import state, Delete
from ravestate_verbaliser import verbaliser
from ravestate.constraint import s

import requests

from reggol import get_logger
logger = get_logger(__name__)

DRQA_SERVER_ADDRESS: str = "drqa_server_address"
ROBOY_ANSWER_SANITY: str = "roboy_answer_sanity"

SERVER_AVAILABLE_CODE = 200


@state(cond=s(":startup"), write="rawio:out")
def hello_world_genqa(ctx):
    server = ctx.conf(key=DRQA_SERVER_ADDRESS)
    if not server:
        logger.error('Server address is not set. Shutting down GenQA.')
        return Delete()
    if not server_up(server):
        return Delete()


@state(cond=s("nlp:is-question"), read="rawio:in", write="rawio:out")
def drqa_module(ctx):
    """
    general question answering using DrQA through a HTTP server
    connection check to server
    post input question and get DrQA answer through the HTTP interface
    depending on the answer score the answer is introduced as sane or insane/unsure :)
    """
    server = ctx.conf(key=DRQA_SERVER_ADDRESS)
    if not server_up(server):
        return Delete(resign=True)
    params = {'question': ctx["rawio:in"]}
    response = requests.get(server, params=params)
    response_json = response.json()
    certainty = response_json["answers"][0]["span_score"]
    # sane answer
    if certainty > ctx.conf(key=ROBOY_ANSWER_SANITY):
        ctx["rawio:out"] = verbaliser.get_random_phrase("question-answering-starting-phrases") + " " + \
                           response_json["answers"][0]["span"]
    # insane/unsure answer
    else:
        ctx["rawio:out"] = verbaliser.get_random_phrase("unsure-question-answering-phrases") \
                           % response_json["answers"][0]["span"] \
                           + "\n" + "Maybe I can find out more if your rephrase the question for me."


def server_up(server):
    status = None
    try:
        status = requests.head(server).status_code
    except requests.exceptions.RequestException or requests.exceptions.ConnectionError:
        logger.error(
            "\n--------"
            "\nThe DrQA server does not seem to be running. GenQA will not work, so no asking general questions!"
            "\nTo set up DrQA, follow the instructions here: https://github.com/roboy/drqa."
            "\n--------")
    return status == SERVER_AVAILABLE_CODE


registry.register(
    name="genqa",
    states=(hello_world_genqa, drqa_module),
    config={
        DRQA_SERVER_ADDRESS: "http://localhost:5000",
        ROBOY_ANSWER_SANITY: 5000
    }
)



# Ravestate static implementations for registering modules


from ravestate import module
import importlib

from reggol import get_logger
logger = get_logger(__name__)

_registered_modules = dict()
_registration_callback = None


def import_module(*, module_name: str, callback):
    """
    Called by context to import a particular ravestate python module.

    * `module_name`: The name of the python module to be imported (must be in pythonpath).

    * `callback`: A callback which should be called when a module calls register() while it is being imported.
    """
    global _registration_callback
    _registration_callback = callback
    importlib.import_module(module_name)
    _registration_callback = None


def register(*, name: str="", props=(), states=(), config=None):
    """
    May be called to register a named set of states, properties and config entries,
    which form a coherent bundle.

    * `name`: The name of the module. Will be prefixed to property and signal names like
     <module name>:<property/signal-name>.

    * `props`: The properties that should be registered.

    * `states`: The states that should be registered.

    * `config`: A dictionary of config entries and their default values, which should be read
     from the default/user config files.
    :return:
    """
    global _registered_modules
    global _registration_callback

    if name in _registered_modules:
        logger.error(f"Attempt to add module {name} twice!")
        return

    if not config:
        config = {}

    _registered_modules[name] = module.Module(name=name, props=props, states=states, config=config)
    if _registration_callback:
        _registration_callback(_registered_modules[name])


def has_module(module_name: str):
    """
    Check whether a module with a particular name has been registered.

    * `module_name`: The name which should be checked for beign registered.

    **Returns:** True if a module with the given name has been registered, false otherwise.
    """
    global _registered_modules
    return module_name in _registered_modules


def get_module(module_name: str):
    """
    Get a registered module with a particular name

    * `module_name`: The name of the moduke which should be retrieved.

    **Returns:** The module with the given name if it was registered, false if otherwise.
    """
    global _registered_modules
    return _registered_modules[module_name] if module_name in _registered_modules else None



# Dummy resource which allowed CausalGroups to track acquisitions
#  for states that don't have any write-props.

from ravestate.property import PropertyBase

from reggol import get_logger
logger = get_logger(__name__)


class Consumable(PropertyBase):
    """
    Dummy resource which allows CausalGroups to track acquisitions
     for states that don't have any write-props.
    """

    def __init__(self, name):
        super().__init__(name=name, allow_read=False, allow_write=False, allow_push=False, allow_pop=False)



from ravestate import registry
from ravestate.state import state, Resign
from ravestate.constraint import s
from ravestate_nlp.question_word import QuestionWord
import ravestate_ontology
from ravestate_verbaliser import verbaliser

from os.path import realpath, dirname, join
import random
import datetime

from reggol import get_logger
logger = get_logger(__name__)

verbaliser.add_folder(join(dirname(realpath(__file__)), "answering_phrases"))

ROBOY_NODE_CONF_KEY = "roboy_node_id"


# TODO: Change this to cond=idle:bored
# @state(cond=s(":startup"), write="rawio:out")
# def hello_world_roboyqa(ctx):
#     ctx["rawio:out"] = "Ask me something about myself!"


@state(cond=s("nlp:contains-roboy") & s("nlp:is-question"), read="nlp:triples", write="rawio:out")
def roboyqa(ctx):
    """
    answers question regarding roboy by retrieving the information out of the neo4j roboy memory graph
    state gets triggered when nlp extracts a new triple: subject, predicate, object
    by analysing the triple the content of the question can be ascertained 
    the necessary information is gathered using the neo4j memory session
    if the triple combination is known and the information could be retrieved an answer will be given

    list of questions that can be answered:
    - who are you?
    - what is your name?
    - how old are you?
    - what is your age?
    - what is your hobby?
    - what are your hobbies?
    - what do you like?
    - where are you from?
    - where do you live?
    - who is your father/dad?
    - who is your brother/sibling?
    - who is your friend?
    - what do you want to become?
    - what are you a member of?
    - what can you do?
    - what are your skills?
    - what have you learned?
    - what are your abilities?
    """
    sess = ravestate_ontology.get_session()
    roboy = sess.retrieve(node_id=ctx.conf(key=ROBOY_NODE_CONF_KEY))[0]
    triple = ctx["nlp:triples"][0]

    category = None
    memory_info = None

    # question word: What?
    if triple.is_question(QuestionWord.OBJECT):
        if triple.match_either_lemma(pred={"like"}, subj={"hobby"}):
            category = "HAS_HOBBY"
        elif triple.match_either_lemma(pred={"learn"}, subj={"skill"}):
            category = "skills"
        elif triple.match_either_lemma(pred={"can"}, subj={"ability"}):
            category = "abilities"
        elif triple.match_either_lemma(subj={"age"}):
            category = "age"
            memory_info = roboy_age(roboy.get_properties(key="birthdate"))
        elif triple.match_either_lemma(subj={"name"}):
            category = "full_name"
            memory_info = roboy.get_properties(key=category)
        elif triple.match_either_lemma(pred={"become"}):
            category = "future"
    # question word: Where?
    elif triple.is_question(QuestionWord.PLACE):
        if triple.match_either_lemma(pred={"be"}):
            category = "FROM"
        elif triple.match_either_lemma(pred={"live"}):
            category = "LIVE_IN"
    # question word: Who?
    elif triple.is_question(QuestionWord.PERSON):
        if triple.match_either_lemma(obj={"father", "dad"}):
            category = "CHILD_OF"
        elif triple.match_either_lemma(obj={"brother", "sibling"}):
            category = "SIBLING_OF"
        elif triple.match_either_lemma(obj={"friend", "girlfriend"}):
            category = "FRIEND_OF"
        else:
            category = "full_name"
            memory_info = roboy.get_properties(key=category)
    elif triple.match_either_lemma(obj={"part", "member"}):
        category = "MEMBER_OF"
    # question word: How?
    elif triple.is_question(QuestionWord.FORM):
        if triple.match_either_lemma(pred={"old"}):
            category = "age"
            memory_info = roboy_age(roboy.get_properties(key="birthdate"))
        elif triple.match_either_lemma(pred={"be"}):
            category = "well_being"
    elif triple.match_either_lemma(obj={"skill"}):
        category = "skills"
    elif triple.match_either_lemma(obj={"ability"}):
        category = "abilities"

    if category and category.isupper() and not isinstance(roboy.get_relationships(key=category), dict):
        node_id = random.sample(roboy.get_relationships(key=category),1)[0]
        memory_info = sess.retrieve(node_id=int(node_id))[0].get_name()

    elif category and category.islower() and not isinstance(roboy.get_properties(key=category), dict):
        property_list = [x.strip() for x in roboy.get_properties(key=category).split(',')]
        memory_info = random.sample(property_list, 1)[0]

    if memory_info:
        ctx["rawio:out"] = verbaliser.get_random_successful_answer(category) % memory_info
    elif category == "well_being":
        ctx["rawio:out"] = verbaliser.get_random_successful_answer(category)
    else:
        return Resign()


def roboy_age(birth_date: str):
    """
    Calculates roboys age given his birth date
    example: birth date = "12.04.18"
    """
    birth_date = datetime.datetime.strptime(birth_date, "%d.%m.%Y")
    today = datetime.datetime.now()
    if today.year > birth_date.year and today.month > birth_date.month:
        age = "%d years" % (today.year - birth_date.year -
                            ((today.month, today.day) < (birth_date.month, birth_date.day)))
    else:
        age = "%d months" % (12 - birth_date.month+today.month)
    return age


registry.register(
    name="roboyqa",
    states=(roboyqa,),
    config={ROBOY_NODE_CONF_KEY: 356}
)


from typing import Set

from ravestate.constraint import s
from ravestate.property import PropertyBase
from ravestate.receptor import receptor
from ravestate.state import state, Delete
from ravestate.wrappers import ContextWrapper
from telegram import Bot, Update, TelegramError
from telegram.ext import Updater, MessageHandler, Filters, Dispatcher

from scientio.ontology.node import Node
from scientio.session import Session
from scientio.ontology.ontology import Ontology

from reggol import get_logger
logger = get_logger(__name__)

import ravestate_rawio
import ravestate_interloc
import ravestate_ontology


TOKEN_CONFIG_KEY: str = "telegram-token"
active_chats: Set[int] = set()
active_users: Set[int] = set()


@state(cond=s(":startup"))
def telegram_run(ctx: ContextWrapper):
    """
    Starts up the telegram bot and adds a handler to write incoming messages to rawio:in
    """

    @receptor(ctx_wrap=ctx, write="rawio:in")
    def telegram_callback(ctx: ContextWrapper, message_text: str):
        """
        Writes the message_text to rawio:in
        """
        ctx["rawio:in"] = message_text

    @receptor(ctx_wrap=ctx, write="interloc:all")
    def push_telegram_interloc(ctx: ContextWrapper, telegram_node: Node, name: str):
        """
        Push the telegram_node into interloc:all:name
        """
        if ctx.push(parentpath="interloc:all", child=PropertyBase(name=name, default_value=telegram_node)):
            logger.debug(f"Pushed {telegram_node} to interloc:all")

    def handle_input(bot: Bot, update: Update):
        """
        Handler for incoming messages
        Adds the chat/user of the incoming message to the set of active_chats/active_users
        Calls the telegram_callback receptor to process the incoming message
        Retrieves scientio Node of User if it exists, otherwise creates it in the scientio session
        Calls the push_telegram_interloc receptor to push the scientio node into interloc:all
        """
        telegram_callback(update.effective_message.text)
        active_chats.add(update.effective_chat.id)

        if update.effective_user not in active_users:
            # set up scientio
            sess: Session = ravestate_ontology.get_session()
            onto: Ontology = ravestate_ontology.get_ontology()

            # create scientio Node of type TelegramPerson
            query = Node(metatype=onto.get_type("TelegramPerson"))
            prop_dict = {'telegram_id': update.effective_user.id}
            if update.effective_user.username:
                prop_dict['name'] = update.effective_user.username
            if update.effective_user.full_name:
                prop_dict['full_name'] = update.effective_user.full_name
            query.set_properties(prop_dict)

            node_list = sess.retrieve(query)
            if not node_list:
                telegram_node = sess.create(query)
                logger.info(f"Created new Node in scientio session: {telegram_node}")
            elif len(node_list) == 1:
                telegram_node = node_list[0]
            else:
                logger.error(f'Found multiple TelegramPersons that matched query: {update.message.chat_id} '
                             f'in scientio session. Cannot push node to interloc:all!')
                return

            # push chat-Node
            push_telegram_interloc(telegram_node, update.effective_chat.id)
            active_users.add(update.effective_user)

    def error(bot: Bot, update: Update, error: TelegramError):
        """
        Log Errors caused by Updates.
        """
        logger.warning('Update "%s" caused error "%s"', update, error)

    """Start the bot."""
    # Create the EventHandler and pass it your bots token.
    token = ctx.conf(key=TOKEN_CONFIG_KEY)
    if not token:
        logger.error('telegram-token is not set. Shutting down telegramio')
        return Delete()

    updater: Updater = Updater(token)
    # Get the dispatcher to register handlers
    dp: Dispatcher = updater.dispatcher
    # handle noncommand-messages with handle_input
    dp.add_handler(MessageHandler(Filters.text, handle_input))
    # log all errors
    dp.add_error_handler(error)
    # Start the Bot
    updater.start_polling()


@state(read="rawio:out")
def telegram_output(ctx: ContextWrapper):
    """
    Sends the content of rawio:out to every currently active chat
    """
    # TODO don't instantiate the updater every time
    token = ctx.conf(key=TOKEN_CONFIG_KEY)
    if not token:
        logger.error('telegram-token is not set. Shutting down telegramio')
        return Delete()

    updater: Updater = Updater(token)
    for chat_id in active_chats:
        updater.bot.send_message(chat_id=chat_id, text=ctx["rawio:out"])



from h5py.h5p import DEFAULT

from ravestate import registry
from ravestate.constraint import s
from ravestate.property import PropertyBase
from ravestate.state import state
from ravestate.receptor import receptor
from ravestate.wrappers import ContextWrapper

import ravestate_rawio
import ravestate_interloc
from ravestate_verbaliser.verbaliser import get_phrase_list
import ravestate_phrases_basic_en
import ravestate_ontology

from scientio.ontology.node import Node
from scientio.session import Session
from scientio.ontology.ontology import Ontology

from reggol import get_logger
logger = get_logger(__name__)

DEFAULT_INTERLOC_ID = "terminal_user"


@state(cond=s(":startup"), read="interloc:all")
def console_input(ctx: ContextWrapper):

    @receptor(ctx_wrap=ctx, write="rawio:in")
    def write_console_input(ctx_input, value: str):
        ctx_input["rawio:in"] = value

    @receptor(ctx_wrap=ctx, write="interloc:all")
    def push_console_interloc(ctx: ContextWrapper, console_node: Node):
        if ctx.push(parentpath="interloc:all", child=PropertyBase(name=DEFAULT_INTERLOC_ID, default_value=console_node)):
            logger.debug(f"Pushed {console_node} to interloc:all")

    @receptor(ctx_wrap=ctx, write="interloc:all")
    def pop_console_interloc(ctx: ContextWrapper):
        if ctx.pop(f"interloc:all:{DEFAULT_INTERLOC_ID}"):
            logger.debug(f"Popped interloc:all:{DEFAULT_INTERLOC_ID}")

    while not ctx.shutting_down():
        input_value = input("> ")
        write_console_input(input_value)

        console_interloc_exists = f"interloc:all:{DEFAULT_INTERLOC_ID}" in ctx.enum("interloc:all")
        # push Node if you got a greeting
        if input_value.strip() in get_phrase_list("greeting") and not console_interloc_exists:
            # set up scientio
            sess: Session = ravestate_ontology.get_session()
            onto: Ontology = ravestate_ontology.get_ontology()

            # create scientio Node of type Person
            query = Node(metatype=onto.get_type("Person"))
            query.set_name("x")
            console_node_list = sess.retrieve(query)
            if not console_node_list:
                console_node = sess.create(query)
                logger.info(f"Created new Node in scientio session: {console_node}")
            elif len(console_node_list) == 1:
                console_node = console_node_list[0]
            else:
                logger.error(f'Found multiple Persons with name {DEFAULT_INTERLOC_ID} in scientio session. Cannot push node to interloc:all!')
                continue

            # push interloc-Node
            push_console_interloc(console_node)

        # pop Node if you got a farewell
        elif input_value.strip() in get_phrase_list("farewells") and console_interloc_exists:
            pop_console_interloc()


@state(read="rawio:out")
def console_output(ctx):
    print(ctx["rawio:out"])


registry.register(
    name="consoleio",
    states=(console_input, console_output)
)



import logging
import os
import random
import yaml
from typing import Dict, List, Optional

from ravestate_verbaliser.qa_phrases import QAPhrases

"""
Produces actual utterances. This should in the future lead to diversifying
the ways Roboy is expressing information.
"""

TYPE_PHRASES: str = "phrases"
TYPE_QA: str = "qa"
phrases: Dict[str, List[str]] = dict()
qa: Dict[str, QAPhrases] = dict()


def add_folder(dirpath: str):
    """
    Parse all files in this folder and add them to the current phrases- and qa-dicts


    * `dirpath`: Path to the folder that should be added
    """
    try:
        for filename in os.listdir(dirpath):
            add_file(os.path.join(dirpath, filename))
    except FileNotFoundError:
        logging.error('Cannot add folder ' + dirpath + ' because it does not exist')


def add_file(path: str):
    """
    Parse the file and add the contents to the current phrases- and qa-dicts


    * `path`: Path to the file that should be added

    **Returns:** True if the file was successfully added, False if there was an error while adding the file
    """
    try:
        with open(path, 'r') as input_file:
            file_data = list(yaml.safe_load_all(input_file))

        for entry in file_data:
            if entry['type'] == TYPE_PHRASES:
                if str(entry['name']) in phrases:
                    logging.error('Import of section ' + str(entry['name']) + ' in ' + path +
                                  ' would overwrite phrases-list entry ' + str(entry['name']))
                phrases[str(entry['name'])] = entry['opts']
            elif entry['type'] == TYPE_QA:
                if str(entry['name']) in qa:
                    logging.error('Import of section ' + str(entry['name']) + ' in ' + path +
                                  ' would overwrite qa-list entry ' + str(entry['name']))
                qa[str(entry['name'])] = QAPhrases(entry)
            else:
                logging.error('Unknown type given for section ' + str(entry['name']) + ' in ' + path)

        return True
    except (FileNotFoundError, KeyError):
        logging.error('Cannot add file ' + path + ' because it does not exist or is in the wrong format.')
        return False


def random_or_none(list_or_none: Optional) -> Optional:
    """
    Get a random element from a list if it is not empty.


    * `list_or_none`: Either a list or None

    **Returns:** random element if given a non-empty list, otherwise None
    """
    return random.choice(list_or_none) if list_or_none else list_or_none


def get_phrase_list(intent: str) -> List[str]:
    """
    Get all imported phrases for this intent as a list


    * `intent`: name-parameter of the yml-section with which the phrases were imported

    **Returns:** None if no phrases are known for this intent, otherwise list of phrases for this intent
    """
    return phrases.get(intent)


def get_random_phrase(intent: str) -> str:
    """
    Get a random phrase for this intent


    * `intent`: name-parameter of the yml-section with which the phrases were imported

    **Returns:** None if no phrases are known for this intent, otherwise a random element of the phrases for this intent
    """
    return random_or_none(phrases.get(intent))


def get_question_list(intent: str) -> List[str]:
    """
    Get all imported questions for this intent as a list


    * `intent`: name-parameter of the yml-section with which the questions were imported

    **Returns:** None if no questions are known for this intent, otherwise list of questions for this intent
    """
    return None if not qa.get(intent) else qa.get(intent).questions


def get_random_question(intent: str) -> str:
    """
    Get a random question for this intent


    * `intent`: name-parameter of the yml-section with which the questions were imported

    **Returns:** None if no questions are known for this intent, otherwise a random element of the questions for this intent
    """
    return random_or_none(get_question_list(intent))


def get_successful_answer_list(intent: str) -> List[str]:
    """
    Get all imported successful answers for this intent as a list


    * `intent`: name-parameter of the yml-section with which the successful answers were imported

    **Returns:** None if no successful answers are known for this intent, otherwise list of successful answers for this intent
    """
    return None if not qa.get(intent) else qa.get(intent).successful_answers


def get_random_successful_answer(intent: str) -> str:
    """
    Get a random successful answer for this intent


    * `intent`: name-parameter of the yml-section with which the successful answers were imported

    **Returns:** None if no successful answers are known for this intent,
        otherwise a random element of the successful answers for this intent
    """
    return random_or_none(get_successful_answer_list(intent))


def get_failure_answer_list(intent: str) -> List[str]:
    """
    Get all imported failure answers for this intent as a list


    * `intent`: name-parameter of the yml-section with which the failure answers were imported

    **Returns:** None if no failure answers are known for this intent, otherwise list of failure answers for this intent
    """
    return None if not qa.get(intent) else qa.get(intent).failure_answers


def get_random_failure_answer(intent: str) -> str:
    """
    Get a random failure answer for this intent


    * `intent`: name-parameter of the yml-section with which the failure answers were imported

    **Returns:** None if no failure answers are known for this intent,
        otherwise a random element of the failure answers for this intent
    """
    return random_or_none(get_failure_answer_list(intent))


def get_followup_question_list(intent: str) -> List[str]:
    """
    Get all imported followup questions for this intent as a list


    * `intent`: name-parameter of the yml-section with which the followup questions were imported

    **Returns:** None if no followup questions are known for this intent, otherwise list of followup questions for this intent
    """
    return None if not qa.get(intent) else qa.get(intent).followup_questions


def get_random_followup_question(intent: str) -> str:
    """
    Get a random followup question for this intent


    * `intent`: name-parameter of the yml-section with which the followup questions were imported

    **Returns:** None if no followup questions are known for this intent, 
        otherwise a random element of the followup questions for this intent
    """
    return random_or_none(get_followup_question_list(intent))


def get_followup_answer_list(intent: str) -> List[str]:
    """
    Get all imported followup answers for this intent as a list


    * `intent`: name-parameter of the yml-section with which the followup answers were imported

    **Returns:** None if no followup answers are known for this intent, otherwise list of followup answers for this intent
    """
    return None if not qa.get(intent) else qa.get(intent).followup_answers


def get_random_followup_answer(intent: str) -> str:
    """
    Get a random followup answer for this intent


    * `intent`: name-parameter of the yml-section with which the followup answers were imported

    **Returns:** None if no followup answers are known for this intent,
        otherwise a random element of the followup answers for this intent
    """
    return random_or_none(get_followup_answer_list(intent))



from ravestate import registry
from ravestate.state import state
from ravestate.constraint import s

from scientio.ontology.ontology import Ontology
from scientio.session import Session

from os.path import realpath, dirname, join

from ravestate_ontology.dummy_session import DummySession

from reggol import get_logger
logger = get_logger(__name__)

onto = None
sess = None
NEO4J_ADDRESS_KEY: str = "neo4j_address"
NEO4J_USERNAME_KEY: str = "neo4j_username"
NEO4J_PASSWORD_KEY: str = "neo4j_pw"


@state(cond=s(":startup"))
def hello_world_ontology(ctx):
    """
    Creates a scientio session with neo4j backend.
    (neo4j is a knowledge graph)
    an ontology is loaded from a yaml file - it can be accessed with a getter
    (the ontology description is a collection of named entity types,
    their properties and relationships)
    then the session is created - it can be accessed with a getter
    """
    # TODO: Make sess and onto context properties?
    global onto, sess
    onto = Ontology(path_to_yaml=join(dirname(realpath(__file__)), "ravestate_ontology.yml"))
    # Create a session (with default Neo4j backend)
    try:
        sess = Session(
            ontology=onto,
            neo4j_address=ctx.conf(key=NEO4J_ADDRESS_KEY),
            neo4j_username=ctx.conf(key=NEO4J_USERNAME_KEY),
            neo4j_password=ctx.conf(key=NEO4J_PASSWORD_KEY))
    except Exception:
        logger.error(
            "\n--------"
            f"\nFailed to create scientio session with {ctx.conf(key=NEO4J_ADDRESS_KEY)}!"
            "\nFor more info on how to set up a server, see https://pypi.org/project/scientio."
            "\n--------")
        sess = DummySession()

registry.register(
    name="ontology",
    states=(hello_world_ontology,),
    config={
        NEO4J_ADDRESS_KEY: "bolt://localhost:7687",
        NEO4J_USERNAME_KEY: "neo4j",
        NEO4J_PASSWORD_KEY: "neo4j"
    }
)


def get_session():
    if not sess:
        logger.error("get_session() called before hello_world_ontology was invoked by context.")
        return None
    return sess


def get_ontology():
    if not onto:
        logger.error("get_ontology() called before hello_world_ontology was invoked by context.")
        return None
    return onto


# TODO: Offer ScientioNodeProperty subclass, which enforces that value is a scientio.ontology.Node?!



# Ravestate class which encapsualtes the activation of a single state
import copy

from threading import Thread
from typing import Set, Optional, List, Dict
from collections import defaultdict

from ravestate.icontext import IContext
from ravestate.constraint import Constraint, s
from ravestate.iactivation import IActivation, ISpike
from ravestate.spike import Spike
from ravestate.causal import CausalGroup
from ravestate.state import State, Emit, Delete, Resign, Wipe
from ravestate.wrappers import ContextWrapper

from reggol import get_logger
logger = get_logger(__name__)


class Activation(IActivation):
    """
    Encapsulates the potential activation of a state. Tracks the collection
     of Spikes to fulfill of the state-defined activation constraints.
    """

    # Count how many activations were created per state
    _count_for_state: Dict[State, int] = defaultdict(int)

    death_clock: Optional[int]  # set once pressure() is called
    id: str  # _count_for_state[self.state_to_activate] from ctor time
    name: str
    state_to_activate: State
    constraint: Constraint
    ctx: IContext
    args: List
    kwargs: Dict
    spikes: Set[Spike]
    consenting_causal_groups: Set[CausalGroup]

    def __init__(self, st: State, ctx: IContext):
        self.id = f"{st.name}#{Activation._count_for_state[st]}"
        Activation._count_for_state[st] += 1
        self.name = st.name
        self.state_to_activate = st
        self.constraint = copy.deepcopy(st.constraint_)
        self.ctx = ctx
        self.args = []
        self.kwargs = {}
        self.spikes = set()
        self.consenting_causal_groups = set()
        self.death_clock = None

    def __del__(self):
        logger.debug(f"Deleted {self}")

    def __repr__(self):
        return self.id + (f"[t-{self.death_clock}]" if self.death_clock is not None else "")

    def resources(self) -> Set[str]:
        """
        Return's the set of the activation's write-access property names.
        """
        if self.state_to_activate.write_props:
            return set(self.state_to_activate.write_props)
        else:
            # Return a dummy resource that will be "consumed" by the activation.
            #  This allows CausalGroup to track the spike acquisitions for this
            #  activation, and make sure that a single spike cannot activate
            #  multiple activations for a write-prop-less state.
            return {self.state_to_activate.consumable.fullname()}

    def specificity(self) -> float:
        """
        Returns the lowest specificity among the specificity values of the
         activation's conjunct constraints. The specificity for a single conjunction
         is calculated as the sum of it's component signal's specificities,
         which in turn is calculated as one over the signal's subscriber count.
        """
        return min(
            sum(self.ctx.signal_specificity(sig) for sig in conj.signals())
            for conj in self.constraint.conjunctions())

    def dereference(self, *, spike: Optional[ISpike]=None, reacquire: bool=False, reject: bool=False) -> None:
        """
        Notify the activation, that a single or all spike(s) are not available
         anymore, and should therefore not be referenced anymore by the activation.
        This is called by ... <br>
        ... context when a state is deleted. <br>
        ... causal group, when a referenced signal was consumed for a required property. <br>
        ... causal group, when a referenced signal was wiped. <br>
        ... this activation (with reacquire=True), if it gives in to activation pressure.

        * `spike`: The spike that should be forgotten by the activation, or
         none, if all referenced spikes should be forgotten.

        * `reacquire`: Flag which tells the function, whether for every rejected
         spike, the activation should hook into context for reacquisition
         of a replacement spike.

        * `reject`: Flag which controls, whether de-referenced spikes
         should be explicitely rejected through their causal groups.
        """
        for sig_to_reacquire, dereferenced_instance in self.constraint.dereference(spike):
            if reacquire:
                self.ctx.reacquire(self, sig_to_reacquire)
            if reject and dereferenced_instance:
                with dereferenced_instance.causal_group() as cg:
                    cg.rejected(dereferenced_instance, self, reason=0)

    def acquire(self, spike: Spike) -> bool:
        """
        Let the activation acquire a signal it is registered to be interested in.

        * `spike`: The signal which should fulfill at least one of this activation's
         signal constraints.

        **Returns:** Should return True.
        """
        if self.death_clock is not None:
            self._reset_death_clock()
        return self.constraint.acquire(spike, self)

    def secs_to_ticks(self, seconds: float) -> int:
        """
        Convert seconds to an equivalent integer number of ticks,
         given this activation's tick rate.

        * `seconds`: Seconds to convert to ticks.

        **Returns:** An integer tick count.
        """
        return self.ctx.secs_to_ticks(seconds)

    def pressure(self):
        """
        Called by CausalGroup, to pressure the activation to
         make a decision on whether it is going to retain a reference
         to the given spike, given that there is a lower-
         specificity activation which is ready to run.
        """
        if self.death_clock is None:
            self._reset_death_clock()

    def spiky(self) -> bool:
        """
        Returns true, if the activation has acquired any spikes at all.

        **Returns:** True, if any of this activation's constraint's
         signal is referencing a spike.
        """
        return sum(1 for sig in self.constraint.signals() if sig.spike) > 0

    def update(self) -> bool:
        """
        Called once per tick on this activation, to give it a chance to activate
         itself, or auto-eliminate, or reject spikes which have become too old.

        **Returns:** True, if the target state is activated and teh activation be forgotten,
         false if needs further attention in the form of updates() by context in the future.
        """
        # Update constraint, reacquire for rejected spikes
        for spike in self.constraint.update(self):
            self.ctx.reacquire(self, spike)

        # Iterate over fulfilled conjunctions and look to activate with one of them
        for conjunction in self.constraint.conjunctions():
            if not conjunction.evaluate():
                continue

            # Ask each spike's causal group for activation consent
            spikes_for_conjunct = set((sig.spike, sig.detached) for sig in conjunction.signals())
            consenting_causal_groups = set()
            all_consented = True
            for spike, _ in spikes_for_conjunct:
                cg: CausalGroup = spike.causal_group()
                if cg not in consenting_causal_groups:
                    with cg:
                        if cg.consent(self):
                            consenting_causal_groups.add(cg)
                        else:
                            all_consented = False
                            break  # break spike iteration
            if not all_consented:
                continue

            # Notify all consenting causal groups that activation is going forward
            for cg in consenting_causal_groups:
                with cg:
                    cg.activated(self)

            # Remove references between causal groups <-> self
            for signal in self.constraint.signals():
                if signal.spike:
                    with signal.spike.causal_group() as cg:
                        cg.rejected(signal.spike, self, reason=2)
                    signal.spike = None

            # Withdraw from context for all (unfulfilled) signals (there might
            #  be some unfulfilled conjunctions next to the fulfilled one).
            for sig in self.constraint.signals():
                self.ctx.withdraw(self, sig)

            # Remember spikes/causal-groups for use in activation
            self.spikes = {spike for spike, detached in spikes_for_conjunct if not detached}
            self.consenting_causal_groups = consenting_causal_groups

            # Run activation
            self.run()

            # Do not further iterate over candidate conjunctions
            return True

        # Update auto-elimination countdown
        if self.death_clock is not None:
            if self.death_clock <= 0:
                self.death_clock = None
                self.dereference(reacquire=True, reject=True)
                logger.info(f"Eliminated {self}.")
            else:
                self.death_clock -= 1

        return False

    def run(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        logger.info(f"Activating {self}")
        Thread(target=self._run_private).start()

    def _reset_death_clock(self):
        self.death_clock = self.ctx.lowest_upper_bound_eta(set(
            sig for sig in self.constraint.signals() if not sig.spike))

    def _unique_consenting_causal_groups(self) -> Set[CausalGroup]:
        # if a signal was emitted by this activation, the consenting
        #  causal groups will be merged. this must be respected -> recreate the set.
        return {group for group in self.consenting_causal_groups}

    def _run_private(self):
        context_wrapper = ContextWrapper(ctx=self.ctx, st=self.state_to_activate, spike_parents=self.spikes)

        # Run state function
        result = self.state_to_activate(context_wrapper, *self.args, **self.kwargs)

        # Process state function result
        if isinstance(result, Emit):
            if self.state_to_activate.signal():
                if result.wipe:
                    self.ctx.wipe(self.state_to_activate.signal())
                self.ctx.emit(
                    self.state_to_activate.signal(),
                    parents=self.spikes)
            else:
                logger.error(f"Attempt to emit spike from state {self.name}, which does not specify a signal name!")

        elif isinstance(result, Wipe):
            if self.state_to_activate.signal():
                self.ctx.wipe(self.state_to_activate.signal())
            else:
                logger.error(f"Attempt to wipe spikes from state {self.name}, which does not specify a signal name!")

        elif isinstance(result, Delete):
            self.ctx.rm_state(st=self.state_to_activate)
            if result.resign:
                for cg in self._unique_consenting_causal_groups():
                    with cg:
                        cg.resigned(self)
                return

        elif isinstance(result, Resign):
            for cg in self._unique_consenting_causal_groups():
                with cg:
                    cg.resigned(self)
            return

        # Let participating causal groups know about consumed properties
        for cg in self._unique_consenting_causal_groups():
            with cg:
                cg.consumed(self.resources())



# Interface for the context towards activation and
#  context/property wrappers (needed to preempt cyclic dependencies).


from ravestate import property
from ravestate import state
from typing import Set
from ravestate.constraint import Signal
from ravestate.spike import Spike
from ravestate.iactivation import IActivation


class IContext:

    def __getitem__(self, key):
        """
        Retrieve a property object by name, that was previously added through add_prop()
         by it's full name. The full name is always the combination of the property's
         name and it's parent's name, joined with a colon: For example, if the name
         of a property is `foo` and it belongs to the module `bar` it's full name
         will be `bar:foo`.
        An error message will be generated if no property with the given name was
         added to the context, and None will be returned.

        * `key`: The full name of the property.

        **Returns:** The property object, or None, if no property with the given name
         was added to the context.
        """
        pass

    def add_prop(self, *, prop: property.PropertyBase):
        """
        Add a property to this context. An error message will be generated,
         if a property with the same name has already been added previously.

        * `prop`: The property object that should be added.
        """
        pass

    def rm_prop(self, *, prop: property.PropertyBase):
        """
        Remove a property from this context.
        Generates error message, if the property was not added with add_prop()
         to the context previously.

        * `prop`: The property to remove.object
        """
        pass

    def emit(self, signal: Signal, parents: Set['Spike']=None, wipe: bool=False) -> None:
        """
        Emit a signal to the signal processing loop. Note:
         The signal will only be processed if run() has been called!

        * `signal`: The signal to be emitted.

        * `parents`: The signal's parents, if it is supposed to be integrated into a causal group.

        * `wipe`: Boolean to control, whether wipe(signal) should be called
         before the new spike is created.
        """
        pass

    def wipe(self, signal: Signal):
        """
        Delete all spikes for the given signal. Partially fulfilled states
         that have acquired an affected spike will be forced to reject it.
        Wiping a parent spike will also wipe all child spikes.

        * `signal`: The signal for which all existing spikes (and their children)
         should be invalidated and forgotten.
        """
        pass

    def add_state(self, *, st: state.State):
        """
        Add a state to this context. It will be indexed wrt/ the properties/signals
         it depends on. Error messages will be generated for unknown signals/properties.

        * `st`: The state which should be added to this context.
        """
        pass

    def rm_state(self, *, st: state.State):
        """
        Remove a state from this context. Note, that any state which is constrained
         on the signal that is emitted by the deleted state will also be deleted.

        * `st`: The state to remove. An error message will be generated,
         if the state was not previously added to this context with add_state().
        """
        pass

    def shutting_down(self):
        """
        Retrieve the shutdown flag value, which indicates whether shutdown() has been called.
        """
        pass

    def shutdown(self):
        """
        Sets the shutdown flag and waits for the signal processing thread to join.
        """
        pass

    def conf(self, *, mod, key=None):
        """
        Get a single config value, or all config values for a particular module.

        * `mod`: The module whose configuration should be retrieved.

        * `key`: A specific config key of the given module, if only a single
         config value should be retrieved.

        **Returns:** The value of a single config entry if key and module are both
         specified and valid, or a dictionary of config entries if only the
         module name is specified (and valid).
        """
        pass

    def lowest_upper_bound_eta(self, signals: Set[Signal]) -> int:
        """
        Called by activation when it is pressured to resign. The activation wants
         to know the earliest ETA of one of it's remaining required constraints.

        * `signals`: The signals, whose ETA will be calculated, and among the
         results the minimum ETA will be returned.

        **Returns:** Number of ticks it should take for at least one of the required
         signals to arrive. Fixed value (1) for now.
        """
        pass

    def signal_specificity(self, sig: Signal) -> float:
        """
        Called by state activation to determine it's constraint's specificity.

        * `sig`: The signal whose specificity should be returned.

        **Returns:** The given signal's specificity.
        """
        pass

    def reacquire(self, act: IActivation, sig: Signal):
        """
        Called by activation to go shopping for a new Spike
         for the specified signal, and should for this purpose be referenced by context.

        * `act`: The activation that needs a new spike of the specified nature.

        * `sig`: Signal type for which a new spike is needed.
        """
        pass

    def withdraw(self, act: IActivation, sig: Signal):
        """
        Called by activation to make sure that it isn't referenced
         anymore as looking for the specified signal.
        This might be, because the activation chose to eliminate itself
         due to activation pressure, or because one of the activations
         conjunctions was fulfilled, so it is no longer looking for
         signals to fulfill the remaining conjunctions.

        * `act`: The activation that has lost interest in the specified signal.

        * `sig`: Signal type for which interest is lost.
        """
        pass

    def secs_to_ticks(self, seconds: float) -> int:
        """
        Convert seconds to an equivalent integer number of ticks,
         given this context's tick rate.

        * `seconds`: Seconds to convert to ticks.

        **Returns:** An integer tick count.
        """
        pass



from ravestate.state import state
from ravestate import registry

import ravestate_rawio
import ravestate_verbaliser
import ravestate_phrases_basic_en

from ravestate.constraint import s


@state(cond=s(":startup"), write="verbaliser:intent")
def hello_world(ctx):
    ctx["verbaliser:intent"] = "greeting"


@state(read="rawio:in", write="rawio:out")
def generic_answer(ctx):
    ctx["rawio:out"] = "Your input contains {} characters!".format(len(ctx["rawio:in"]))


@state(read="facerec:face", write="rawio:out")
def face_recognized(ctx):
    if ctx["facerec:face"]:
        ctx["rawio:out"] = "I see you, {}!".format(ctx["facerec:face"])


registry.register(name="hi", states=(hello_world, generic_answer, face_recognized))



from typing import List, Set, Generator, Optional, Tuple
from ravestate.spike import Spike
from ravestate.iactivation import IActivation

from reggol import get_logger
logger = get_logger(__name__)


def s(signal_name: str, *, min_age=0, max_age=5., detached=False):
    """
    Alias to call Signal-constructor

    * `signal_name`: Name of the Signal

    * `min_age`: Minimum age for the signal, in seconds.

    * `max_age`: Maximum age for the signal, in seconds.
     Set to less-than zero for unrestricted age.

    * `detached`: Flag which indicates, whether spikes that fulfill this signal
     are going to have a separate causal group from spikes that
     are generated by a state that uses this signal as a constraint.
    """
    return Signal(signal_name, min_age=min_age, max_age=max_age, detached=detached)


class Constraint:
    """
    Superclass for Signal, Conjunct and Disjunct
    """

    def signals(self) -> Generator['Signal', None, None]:
        logger.error("Don't call this method on the super class Constraint")
        yield None

    def conjunctions(self) -> Generator['Conjunct', None, None]:
        logger.error("Don't call this method on the super class Constraint")
        yield None

    def acquire(self, spike: Spike, act: IActivation):
        logger.error("Don't call this method on the super class Constraint")
        pass

    def evaluate(self) -> bool:
        logger.error("Don't call this method on the super class Constraint")
        return False

    def dereference(self, spike: Optional[Spike]=None) -> Generator[Tuple['Signal', 'Spike'], None, None]:
        logger.error("Don't call this method on the super class Constraint")
        yield None, None

    def update(self, act: IActivation) -> Generator['Signal', None, None]:
        logger.error("Don't call this method on the super class Constraint")
        yield None


class Signal(Constraint):
    """
    Class that represents a Signal
    """
    name: str
    spike: Spike
    min_age: float
    max_age: float
    detached: bool
    _min_age_ticks: int  # written on acquire, when act.secs_to_ticks is available

    def __init__(self, name: str, *, min_age=0., max_age=1., detached=False):
        self.name = name
        # TODO: Convert seconds for min_age/max_age to ticks
        self.min_age = min_age
        self.max_age = max_age
        self.spike = None
        self.detached = detached
        self._min_age_ticks = 0

    def __or__(self, other):
        if isinstance(other, Signal):
            return Disjunct(Conjunct(self), Conjunct(other))
        elif isinstance(other, Conjunct):
            return Disjunct(Conjunct(self), other)
        elif isinstance(other, Disjunct):
            return Disjunct(self, *other)

    def __and__(self, other):
        if isinstance(other, Signal):
            return Conjunct(self, other)
        elif isinstance(other, Conjunct):
            return Conjunct(self, *other)
        elif isinstance(other, Disjunct):
            conjunct_list: List[Conjunct] = []
            for conjunct in other:
                conjunct_list.append(Conjunct(*conjunct, self))
            return Disjunct(*conjunct_list)

    def __eq__(self, other):
        return isinstance(other, Signal) and self.name == other.name

    def __hash__(self):
        return hash(self.name)

    def signals(self) -> Generator['Signal', None, None]:
        yield self

    def conjunctions(self) -> Generator['Conjunct', None, None]:
        yield Conjunct(self)

    def acquire(self, spike: Spike, act: IActivation):
        if not self.spike and self.name == spike.name() and (self.max_age < 0 or spike.age() <= act.secs_to_ticks(self.max_age)):
            self._min_age_ticks = act.secs_to_ticks(self.min_age)
            self.spike = spike
            with spike.causal_group() as cg:
                cg.acquired(spike, act)
            return True
        return False

    def evaluate(self) -> bool:
        return self.spike and self._min_age_ticks <= self.spike.age()

    def dereference(self, spike: Optional[Spike]=None) -> Generator[Tuple['Signal', 'Spike'], None, None]:
        if (not spike and self.spike) or (spike and self.spike is spike):
            former_signal_instance = self.spike
            self.spike = None
            yield self, former_signal_instance

    def update(self, act: IActivation) -> Generator['Signal', None, None]:
        # Reject spike, once it has become too old
        if self.spike and self.max_age >= 0 and self.spike.age() > act.secs_to_ticks(self.max_age):
            with self.spike.causal_group() as cg:
                cg.rejected(self.spike, act, reason=1)
                self.spike = None
                yield self

    def __str__(self):
        return self.name


class Conjunct(Constraint):
    """
    Class that represents a Conjunction of Signals
    """
    _signals: Set[Signal]
    _hash: Tuple[str]

    def __init__(self, *args):
        for arg in args:
            if not isinstance(arg, Signal):
                logger.error("Conjunct can only be constructed with Signals.")
                raise ValueError
        self._signals = set(args)
        self._hash = hash(tuple(sorted(sig.name for sig in self._signals)))

    def __iter__(self):
        for signal in self._signals:
            yield signal

    def __or__(self, other):
        if isinstance(other, Signal):
            return Disjunct(self, Conjunct(other))
        elif isinstance(other, Conjunct):
            return Disjunct(self, other)
        elif isinstance(other, Disjunct):
            return Disjunct(self, *other)

    def __and__(self, other):
        if isinstance(other, Signal):
            return Conjunct(other, *self)
        elif isinstance(other, Conjunct):
            return Conjunct(*self, *other)
        elif isinstance(other, Disjunct):
            conjunct_list: List[Conjunct] = []
            for conjunct in other:
                conjunct_list.append(Conjunct(*conjunct, *self))
            return Disjunct(*conjunct_list)

    def __contains__(self, item):
        return item in self._signals

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        return isinstance(other, Conjunct) and self._hash == other._hash

    def signals(self) -> Generator['Signal', None, None]:
        return (sig for sig in self._signals)

    def conjunctions(self) -> Generator['Conjunct', None, None]:
        yield self

    def acquire(self, spike: Spike, act: IActivation):
        result = False
        for si in self._signals:
            result |= si.acquire(spike, act)
        return result

    def evaluate(self) -> bool:
        return all(map(lambda si: si.evaluate(), self._signals))

    def dereference(self, spike: Optional[Spike]=None) -> Generator[Tuple['Signal', 'Spike'], None, None]:
        return (result for child in self._signals for result in child.dereference(spike))

    def update(self, act: IActivation) -> Generator['Signal', None, None]:
        return (result for child in self._signals for result in child.update(act))

    def __str__(self):
        return "(" + " & ".join(map(lambda si: si.__str__(), self._signals)) + ")"


class Disjunct(Constraint):
    """
    Class that represents a Disjunction of Conjunctions
    """
    _conjunctions: Set[Conjunct]

    def __init__(self, *args):
        for arg in args:
            if not isinstance(arg, Conjunct):
                logger.error("Disjunct can only be constructed with conjuncts.")
                raise ValueError
        self._conjunctions = set(args)

    def __iter__(self):
        for conjunct in self._conjunctions:
            yield conjunct

    def __or__(self, other):
        if isinstance(other, Signal):
            return Disjunct(*self, Conjunct(other))
        elif isinstance(other, Conjunct):
            return Disjunct(*self, other)
        elif isinstance(other, Disjunct):
            return Disjunct(*self, *other)

    def __and__(self, other):
        if isinstance(other, Signal):
            conjunct_list: List[Conjunct] = []
            for conjunct in self:
                conjunct_list.append(Conjunct(*conjunct, other))
            return Disjunct(*conjunct_list)
        elif isinstance(other, Conjunct):
            conjunct_list: List[Conjunct] = []
            for conjunct in self:
                conjunct_list.append(Conjunct(*conjunct, *other))
            return Disjunct(*conjunct_list)
        elif isinstance(other, Disjunct):
            logger.error("Can't conjunct two disjunctions.")
            raise ValueError("Can't conjunct two disjunctions.")

    def signals(self) -> Generator['Signal', None, None]:
        return (signal for conjunct in self._conjunctions for signal in conjunct._signals)

    def conjunctions(self) -> Generator['Conjunct', None, None]:
        return (conj for conj in self._conjunctions)

    def acquire(self, spike: Spike, act: IActivation):
        result = False
        for conjunct in self._conjunctions:
            result |= conjunct.acquire(spike, act)
        return result

    def evaluate(self) -> bool:
        return any(map(lambda si: si.evaluate(), self._conjunctions))

    def dereference(self, spike: Optional[Spike]=None) -> Generator[Tuple['Signal', 'Spike'], None, None]:
        return (result for child in self._conjunctions for result in child.dereference(spike))

    def update(self, act: IActivation) -> Generator['Signal', None, None]:
        return (result for child in self._conjunctions for result in child.update(act))

    def __str__(self):
        return " | ".join(map(lambda conjunct: conjunct.__str__(), self._conjunctions))



# Ravestate class which represents a cluster of module configurations
# for a single context instance.

from ravestate import module
from collections import defaultdict
from typing import List, Any

import yaml
from yamlinclude import YamlIncludeConstructor

YamlIncludeConstructor.add_to_loader_class(yaml.SafeLoader)

from reggol import get_logger
logger = get_logger(__name__)


class Configuration:
    """
    The Configuration class maintains a dictionary of key-value stores, which
    represent configuration entries for specific named modules. The key-value
    stores may be successively updated with consecutive yaml files, where each
    yaml document has the following content:

      ---
      module: module-name
      config:
        key-a: value-a
        key-b: value-b
        # etc
      ---
    """

    def __init__(self, paths: List[str]):
        """
        Constructor for a Configuration instance.

        * `paths`: Yaml file paths, where each file contains zero or more
        yaml documents according to the layout described above.
        """
        self.parsed_config_per_module = defaultdict(lambda: {})
        self.config_per_module = {}
        for path in paths:
            self.read(path)

    def add_conf(self, mod: module.Module):
        """
        Register a set of allowed config entries for a specific module.
        Correctly typed values for allowed keys, that were previously
        parsed during construction from the yaml files, will be applied
        immediately.

        * `mod`: A module object with a name and a conf dict.
        """
        if mod.name in self.config_per_module:
            logger.warning(f"add_conf called repeatedly for module name f{mod.name}!")
        self.config_per_module[mod.name] = mod.conf
        self._apply_parsed_config(mod.name)

    def get_conf(self, module_name: str):
        """
        Retrieve updated config values for a module that was previously
        registered with add_conf.

        * `module_name`: The module name for which configuration should be retrieved.

        **Returns:** A dictionary which contains exactly the keys that were contained in
        the module configuration dictionary during add_conf, or an empty dictionary
        if the module name is unknown.
        """
        if module_name not in self.parsed_config_per_module:
            logger.error(f"Bad request for unknown module config {module_name}!")
            return {}
        return self.config_per_module[module_name]

    def get(self, module_name: str, key: str) -> Any:
        """
        Gte the current value of a config entry.

        * `module_name`: The module that provides the config entry.

        * `key`: A config key for the module that was previously added through add_conf.

        **Returns:** The current value, or None, if the entry does not exist.
        """
        if module_name not in self.config_per_module:
            logger.error(f"Attempt to run get() for unknown modname {module_name}!")
            return None
        target_conf = self.config_per_module[module_name]
        if key not in target_conf:
            logger.warning(f"Cannot read unknown conf key {key} for module {module_name}.")
            return None
        return target_conf[key]

    def set(self, module_name: str, key: str, value: Any):
        """
        Set the current value of a config entry.

        * `module_name`: The module of the config entry.

        * `key`: A config key for the module that was previously added through add_conf.

        * `value`: The new value for the config entry. An error will be raised,
         if the type of the new value does not match the type of the old value.
        """
        if module_name not in self.config_per_module:
            logger.error(f"Attempt to run set() for unknown modname {module_name}!")
            return
        target_conf = self.config_per_module[module_name]
        if key not in target_conf:
            logger.warning(f"Not setting unknown conf key {key} for module {module_name}.")
            return
        current_value = target_conf[key]
        if isinstance(current_value, list):
            if not isinstance(value, list):
                value = [value]
            if len(current_value) and len(value) and not isinstance(value[0], type(current_value[0])):
                value = [type(current_value[0])(item) for item in value]
        if not isinstance(value, type(current_value)):
            logger.warning(
                f"Config entry for {module_name}.{key} has conflicting type {type(value).__name__} " +
                f"(should be {type(current_value).__name__}).")
        target_conf[key] = value

    def write(self, path: str):
        """
        Write all current config entries to a yaml file.

        * `path`: The file path to write. Will be overwritten!
        """
        with open(path, mode='w') as file:
            yaml.safe_dump_all((
                {"module": module_name, "config": config_entries}
                for module_name, config_entries in self.config_per_module),
                file,
                default_flow_style=False)

    def read(self, path: str):
        """
        Loads all documents from a yaml file and tries to interpret them
        as configuration objects as described above.

        * `path`: The yaml file path from which to load config documents.
        """
        with open(path, mode='r') as file:
            try:
                configs = yaml.safe_load_all(file)
            except Exception as e:
                logger.warning(f"Could not load config file {path}")
                return
            for config in configs:
                if not isinstance(config, dict) or "module" not in config or "config" not in config:
                    logger.warning(f"Skipping invalid entry for config file {path}.")
                    continue
                module_name = config["module"]
                conf = config["config"]
                if not isinstance(conf, dict):
                    logger.warning(f"Skipping invalid config for {module_name} in {path}.")
                    continue
                target_conf = self.parsed_config_per_module[module_name]
                target_conf.update(conf)

    def _apply_parsed_config(self, module_name: str):
        """
        Apply loaded config values to the actual configuration of a specific module.
        Will be called when add_conf is called with a particular module.
        :param module_name:
        :return:
        """
        if module_name not in self.parsed_config_per_module:
            return
        if module_name not in self.config_per_module:
            logger.error(f"Attempt to run _apply_parsed_config for unknown modname {module_name}!")
            return

        config = self.parsed_config_per_module[module_name]

        for key, value in config.items():
            self.set(module_name, key, value)





from ravestate import registry
from ravestate.property import PropertyBase


registry.register(
    name="rawio",
    props=(
        PropertyBase(name="in", default_value="", allow_pop=False, allow_push=False),
        PropertyBase(name="out", default_value="", allow_pop=False, allow_push=False))
)



# Ravestate module class

from typing import Dict, Any, Tuple
from ravestate.property import PropertyBase
from ravestate.state import State


class Module:
    """
    Atomic class, which encapsulates a named set of states, properties and config entries,
    which form a coherent bundle.
    """

    def __init__(self, *, name: str,
                 props: Tuple[PropertyBase]=(),
                 states: Tuple[State]=(),
                 config: Dict[str, Any]=None):

        if not isinstance(props, tuple):
            props = (props,)
        if not isinstance(states, tuple):
            states = (states,)
        if not config:
            config = {}

        self.name = name
        self.props = props
        self.states = states
        self.conf = config

        for prop in props:
            prop.set_parent_path(name)
        for st in states:
            st.module_name = name



# Ravestate wrapper classes which limit a state's context access

from ravestate.property import PropertyBase
from ravestate import state
from ravestate import icontext
from ravestate.spike import Spike
from typing import Any, Generator, Set

from ravestate.constraint import s

from reggol import get_logger
logger = get_logger(__name__)


class PropertyWrapper:
    """
    Encapsulates a property, and annotates it with additional r/w perms and a context.
    The context is used to trigger the proper :changed, :pushed, :popped, :deleted
    signals when the property is accessed. The wrapper also takes care of locking the property
    when it is supposed to be written to, and freezing the property's value if it is supposed to
    be read from.
    """
    def __init__(self, *,
                 spike_parents: Set[Spike] = None,
                 prop: PropertyBase,
                 ctx: icontext.IContext,
                 allow_read: bool,
                 allow_write: bool):
        self.prop = prop
        self.ctx = ctx
        self.allow_read = allow_read and prop.allow_read
        self.allow_write = allow_write and (prop.allow_write | prop.allow_push | prop.allow_pop)
        self.frozen_value = None
        self.spike_parents = spike_parents

        self.prop.lock()
        if self.allow_read:
            self.frozen_value = prop.value
        if not self.allow_write:
            self.prop.unlock()

    def __del__(self):
        if self.allow_write:
            self.prop.unlock()

    def get(self) -> Any:
        """
        Read the current property value or the value of children of the property if child-param is given

        * `child`: top-down list of child ancestry of the child to get the value from
        """
        if not self.allow_read:
            logger.error(f"Unauthorized read access in property-wrapper for {self.prop.fullname()}!")
            return None
        elif self.allow_write:
            return self.prop.read()
        return self.frozen_value

    def set(self, value: Any):
        """
        Write a new value to the property.

        * `value`: The new value.

        **Returns:** True if the value has changed and :changed should be signaled, false otherwise.
        """
        if not self.allow_write:
            logger.error(f"Unauthorized write access in property-wrapper {self.prop.fullname()}!")
            return False
        if self.prop.write(value):
            self.ctx.emit(self.prop.changed_signal(), parents=self.spike_parents, wipe=True)
            return True
        return False

    def push(self, child: PropertyBase):
        """
        Add a child to the property or to children of the property

        * `child`: Parent-less, child-less property object to add.
         Name of the child must be unique among existing children of this property.

        **Returns:** True if the push was successful, False otherwise
        """
        if not self.allow_write:
            logger.error(f"Unauthorized push access in property-wrapper {self.prop.fullname()}!")
            return False
        if self.prop.push(child):
            self.ctx.emit(self.prop.pushed_signal(), parents=self.spike_parents, wipe=True)
            return True
        return False

    def pop(self, childname: str):
        """
        Remove a child from the property or from children of the property

        * `childname`: Name of the direct child to be removed

        **Returns:** True if the pop was successful, False otherwise
        """
        if not self.allow_write:
            logger.error(f"Unauthorized pop access in property-wrapper {self.prop.fullname()}!")
            return False
        if self.prop.pop(childname):
            self.ctx.emit(self.prop.popped_signal(), parents=self.spike_parents, wipe=True)
            return True
        return False

    def enum(self) -> Generator[str, None, None]:
        """
        Get the full paths of each of this property's children.
        """
        if not self.allow_read:
            logger.error(f"Unauthorized read access in property-wrapper for {self.prop.fullname()}!")
            return (_ for _ in ())
        return (child.fullname() for _, child in self.prop.children.items())


class ContextWrapper:
    """
    Encapsulates a context towards a state, only offering properties with permissions
    as declared by the state beforehand.
    """

    def __init__(self, *, ctx: icontext.IContext, st: state.State, spike_parents: Set[Spike]=None):
        self.st = st
        self.ctx = ctx
        self.properties = dict()
        self.spike_parents = spike_parents
        # Recursively complete properties dict with children:
        for propname in st.write_props + st.read_props:
            # May have been covered by a parent before
            if propname not in self.properties:
                prop_and_children = ctx[propname].gather_children()
                for prop in prop_and_children:
                    # Child may have been covered by a parent before
                    if prop.fullname() not in self.properties:
                        self.properties[prop.fullname()] = PropertyWrapper(
                            prop=prop, ctx=ctx,
                            spike_parents=self.spike_parents,
                            allow_read=propname in st.read_props,
                            allow_write=propname in st.write_props)

    def __setitem__(self, key, value):
        if key in self.properties:
            return self.properties[key].set(value)
        else:
            logger.error(f"State {self.st.name} attempted to write property {key} without permission!")

    def __getitem__(self, key) -> Any:
        if key in self.properties:
            return self.properties[key].get()
        else:
            logger.error(f"State {self.st.name} attempted to access property {key} without permission!")

    def add_state(self, st: state.State):
        self.ctx.add_state(st=st)

    def shutdown(self):
        self.ctx.shutdown()

    def shutting_down(self):
        return self.ctx.shutting_down()

    def conf(self, *, mod=None, key=None):
        if not mod:
            mod = self.st.module_name
        return self.ctx.conf(mod=mod, key=key)

    def push(self, parentpath: str, child: PropertyBase):
        """
        Add a child to a property.
         Note: Child must not yet have a parent or children of itself.
          Write-access to parent is needed.

        * `parentpath`: Path of the parent that should receive the new child.

        * `child`: Parent-less, child-less property object to add.

        **Returns:** True if the push was successful, False otherwise
        """
        if child.parent_path:
            logger.error(f"State {self.st.name} attempted to push child property {child.name} to parent {parentpath}, but it already has parent {child.parent_path}!")
            return False
        if parentpath in self.properties:
            if self.properties[parentpath].push(child):
                self.properties[child.fullname()] = PropertyWrapper(
                    prop=child, ctx=self.ctx,
                    spike_parents=self.spike_parents,
                    allow_read=self.properties[parentpath].allow_read,
                    allow_write=self.properties[parentpath].allow_write)
                self.ctx.add_prop(prop=child)
                return True
        else:
            logger.error(f'State {self.st.name} attempted to add child-property {child.name} to non-accessible parent {parentpath}!')
            return False

    def pop(self, path: str):
        """
        Delete a property (remove it from context and it's parent).
         Note: Write-access to parent is needed!

        * `path`: Path to the property. Must be nested (not root-level)!

        **Returns:** True if the pop was successful, False otherwise
        """
        path_parts = path.split(":")
        if len(path_parts) < 3:
            logger.error(f"State {self.st.name}: Path to pop is not a nested property: {path}")
            return False
        parentpath = ":".join(path_parts[:-1])
        if parentpath in self.properties:
            if self.properties[parentpath].pop(path_parts[-1]):
                self.ctx.rm_prop(prop=self.properties[path].prop)
                # Remove property from own dict
                del self.properties[path]
                # Also remove the deleted propertie's children
                for childpath in list(self.properties.keys()):
                    if childpath.startswith(path+":"):
                        self.ctx.rm_prop(prop=self.properties[childpath].prop)
                        del self.properties[childpath]
                return True
            else:
                logger.error(f'State {self.st.name} attempted to remove non-existent child-property {path}')
                return False
        else:
            logger.error(f'State {self.st.name} attempted to remove child-property {path} from non-existent parent-property {parentpath}')
            return False

    def enum(self, path) -> Generator[str, None, None]:
        """
        Enumerate a propertie's children by their full pathes.
        """
        if path in self.properties:
            return self.properties[path].enum()
        else:
            logger.error(f"State {self.st.name} attempted to enumerate property {path} without permission!")




from typing import List


def check_if_list(value):
    if isinstance(value, List):
        return value
    raise AttributeError("Value is not a list: " + str(value))


class QAPhrases:
    """
    Getting values from a yml file
    Parses files containing predefined questions and answers
    Expects the following input pattern:
    ---
    type: qa
    name: "INTENT"
    Q:
    - "Question phrasing 1"
    - "Question phrasing 2"
    - "Question phrasing 3"
    A:
      SUCCESS:
      - "Possible answer on success 1"
      - "Possible answer on success 2"
      FAILURE:
      - "Possible answer on failure"
    FUP:
      Q:
      - "Possible follow up question"
      A:
      - "Possible follow up answer"

    See more examples in resources / sentences
    """
    questions: List[str] = None
    successful_answers: List[str] = None
    failure_answers: List[str] = None
    followup_questions: List[str] = None
    followup_answers: List[str] = None

    def __init__(self, yaml_data: dict):
        """
        Initializes the lists for the different QA-elements with the data in the yaml_data-dict


        * `yaml_data`: dict of one section of a yml file with the structure of the above example
            converted to a dict by yaml.safe_load_all(path)
        """

        for key in yaml_data:
            if key == 'Q':
                self.questions = check_if_list(yaml_data[key])
            elif key == 'A':
                for subkey in yaml_data[key]:
                    if subkey == 'SUCCESS':
                        self.successful_answers = check_if_list(yaml_data[key][subkey])
                    elif subkey == 'FAILURE':
                        self.failure_answers = check_if_list(yaml_data[key][subkey])
            elif key == 'FUP':
                for subkey in yaml_data[key]:
                    if subkey == 'Q':
                        self.followup_questions = check_if_list(yaml_data[key][subkey])
                    elif subkey == 'A':
                        self.followup_answers = check_if_list(yaml_data[key][subkey])



from scientio.ontology.node import Node


class DummySession:
    """
    Dummy Session that will be used, when Neo4j is unavailable.
    """

    def __init__(self, *args, **kwargs):
        pass

    def create(self, request: Node):
        return request

    def retrieve(self, *args, **kwargs):
        return []

    def update(self, request: Node) -> Node:
        return request

    def delete(self, *args, **kwargs) -> bool:
        return False


import pytest
from reggol import strip_prefix
from testfixtures import LogCapture

from ravestate.constraint import s
from ravestate.context import Context
from ravestate.property import PropertyBase
from ravestate.state import State, state
from ravestate.wrappers import PropertyWrapper, ContextWrapper
from ravestate.activation import Activation

DEFAULT_MODULE_NAME = 'module'
DEFAULT_PROPERTY_NAME = 'property'
DEFAULT_PROPERTY_FULLNAME = f"{DEFAULT_MODULE_NAME}:{DEFAULT_PROPERTY_NAME}"
DEFAULT_PROPERTY_VALUE = 'Kruder'
DEFAULT_PROPERTY_CHANGED = f"{DEFAULT_PROPERTY_FULLNAME}:changed"
NEW_PROPERTY_VALUE = 'Dorfmeister'


@pytest.fixture
def state_fixture(mocker):
    @state(write=(DEFAULT_PROPERTY_FULLNAME,), read=(DEFAULT_PROPERTY_FULLNAME,))
    def state_mock_fn(ctx):
        pass
    state_mock_fn.module_name = DEFAULT_MODULE_NAME
    return state_mock_fn


@pytest.fixture
def state_signal_a_fixture(mocker):
    @state(read=(DEFAULT_PROPERTY_FULLNAME,), signal_name="a")
    def state_mock_a_fn(ctx):
        pass
    state_mock_a_fn.module_name = DEFAULT_MODULE_NAME
    return state_mock_a_fn


@pytest.fixture
def state_signal_b_fixture(mocker):
    @state(signal_name="b", cond=s("module:a"))
    def state_mock_b_fn(ctx):
        pass
    state_mock_b_fn.module_name = DEFAULT_MODULE_NAME
    return state_mock_b_fn


@pytest.fixture
def state_signal_c_fixture(mocker):
    @state(signal_name="c", cond=s("module:a"))
    def state_mock_c_fn(ctx):
        pass
    state_mock_c_fn.module_name = DEFAULT_MODULE_NAME
    return state_mock_c_fn


@pytest.fixture
def state_signal_d_fixture(mocker):
    @state(signal_name="d", cond=s("module:b")|s("module:c"))
    def state_mock_c_fn(ctx):
        pass
    state_mock_c_fn.module_name = DEFAULT_MODULE_NAME
    return state_mock_c_fn


@pytest.fixture
def context_fixture(mocker):
    return Context()


@pytest.fixture
def context_with_property_fixture(mocker, context_fixture) -> Context:
    prop = PropertyBase(name=DEFAULT_PROPERTY_NAME, default_value=DEFAULT_PROPERTY_VALUE)
    prop.set_parent_path(DEFAULT_MODULE_NAME)
    context_fixture.add_prop(prop=prop)
    mocker.patch.object(context_fixture, 'add_prop')
    return context_fixture


@pytest.fixture
def context_with_property_and_state_fixture(mocker, context_with_property_fixture, state_fixture):
    context_with_property_fixture.add_state(st=state_fixture)
    mocker.patch.object(context_with_property_fixture, 'add_state')
    return context_with_property_fixture


@pytest.fixture
def context_wrapper_fixture(context_with_property_fixture, state_fixture):
    return ContextWrapper(ctx=context_with_property_fixture, st=state_fixture)


@pytest.fixture
def activation_fixture(state_fixture: State, context_with_property_and_state_fixture: Context):
    return Activation(state_fixture, context_with_property_and_state_fixture)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

required_url = []
required = []
with open("requirements.txt", "r") as freq:
    for line in freq.read().split():
        if "://" in line:
            required_url.append(line)
        else:
            required.append(line)

packages = setuptools.find_packages("modules", exclude=["reggol*"])

setuptools.setup(
    name="ravestate",
    version="0.3.post",
    url="https://github.com/roboy/ravestate",
    author="Roboy",
    author_email="info@roboy.org",

    description="Ravestate is a reactive library for real-time natural language dialog systems.",
    long_description=long_description,
    long_description_content_type="text/markdown",

    package_dir={package: 'modules/'+package for package in packages},
    packages=packages,
    include_package_data=True,
    package_data={
        'ravestate_phrases_basic_en': ['en/*.yml'],
        'ravestate_ontology': ['ravestate_ontology.yml'],
        'ravestate_roboyqa': ['answering_phrases/RoboyInfoList.yml']
    },
    scripts=["rasta"],

    install_requires=required + ["reggol"],
    dependency_links=required_url,
    python_requires='>=3.7',

    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)



from ravestate.property import PropertyBase
from ravestate import registry

# TODO: Make interloc:all a special property type, that only accepts ScientioNodeProperty as children
registry.register(
    name="interloc",
    props=PropertyBase(name="all", allow_read=True, allow_write=False, allow_push=True, allow_pop=True),
)



import logging

from ravestate import registry
from ravestate.property import PropertyBase
from ravestate.state import state
from ravestate_verbaliser import verbaliser


@state(read="verbaliser:intent", write="rawio:out")
def react_to_intent(ctx):
    """
    Looks for intents written to the verbaliser:intent property and
    writes a random phrase for that intent to rawio:out
    """
    intent = ctx["verbaliser:intent"]
    phrase = verbaliser.get_random_phrase(intent)
    if phrase:
        ctx["rawio:out"] = phrase
    else:
        logging.error('No phrase for intent ' + intent + ' found.')


registry.register(
    name="verbaliser",
    states=(react_to_intent,),
    props=(PropertyBase(name="intent", default_value="", allow_push=False, allow_pop=False),))



# Ravestate context class

from threading import Thread, Lock, Event
from typing import Optional, Any, Tuple, Set, Dict, Iterable, List
from collections import defaultdict
from math import ceil
from copy import deepcopy
import gc

from ravestate.icontext import IContext
from ravestate.module import Module
from ravestate.state import State
from ravestate.property import PropertyBase
from ravestate.iactivation import IActivation
from ravestate.activation import Activation
from ravestate import registry
from ravestate import argparse
from ravestate.config import Configuration
from ravestate.constraint import s, Signal, Conjunct, Disjunct
from ravestate.spike import Spike

from reggol import get_logger
logger = get_logger(__name__)


class Context(IContext):

    _default_signal_names: Tuple[str] = (":startup", ":shutdown", ":idle")

    _core_module_name = "core"
    _import_modules_config = "import"
    _tick_rate_config = "tickrate"

    _lock: Lock

    _properties: Dict[str, PropertyBase]
    _spikes: Dict[Spike, bool]  # Bool says not-wiped (acquisition allowed) true/false

    # Some activations that have all constraints fulfilled
    #  still need to be updated, because they are waiting for
    #  their turn. While they wait, they are stored here.
    # Also, this is just a pretty fucking useful index.
    _activations_per_state: Dict[State, Set[Activation]]

    # This is the bar part of the gaybar - here, activations
    #  register for certain signal spikes which they still need to fulfill.
    _act_per_state_per_signal_age: Dict[
        Signal,
        Dict[
            int,
            Dict[
                State,
                Set[Activation]
            ]
        ]
    ]

    # This data structure is used to complete state constraints:
    #  If a state depends on a signal X that is an effect of signals
    #  Y or Z, the constraint will become (Y & X) | (Z & X).
    _signal_causes: Dict[Signal, List[Conjunct]]

    _config: Configuration
    _core_config: Dict[str, Any]
    _run_task: Thread
    _shutdown_flag: Event

    def __init__(self, *arguments):
        """
        Construct a context from command line arguments.

        * `arguments`: A series of command line arguments which can be parsed
         by the ravestate command line parser (see argparse.py).
        """
        modules, overrides, config_files = argparse.handle_args(*arguments)
        self._config = Configuration(config_files)
        self._core_config = {
            self._import_modules_config: [],
            self._tick_rate_config: 20
        }
        self._config.add_conf(Module(name=self._core_module_name, config=self._core_config))
        self._lock = Lock()
        self._shutdown_flag = Event()
        self._properties = dict()
        self._spikes = defaultdict(lambda: True)
        self._act_per_state_per_signal_age = dict()
        self._signal_causes = dict()
        self._activations_per_state = dict()
        self._run_task = None

        # Register default signals
        for signame in self._default_signal_names:
            self._add_sig(s(signame))

        # Set required config overrides
        for module_name, key, value in overrides:
            self._config.set(module_name, key, value)
        if self._core_config[self._tick_rate_config] < 1:
            logger.error("Attempt to set core config `tickrate` to a value less-than 1!")
            self._core_config[self._tick_rate_config] = 1

        # Load required modules
        for module_name in self._core_config[self._import_modules_config]+modules:
            self.add_module(module_name)

    def emit(self, signal: Signal, parents: Set[Spike]=None, wipe: bool=False) -> None:
        """
        Emit a signal to the signal processing loop. Note:
         The signal will only be processed if run() has been called!

        * `signal`: The signal to be emitted.

        * `parents`: The signal's parents, if it is supposed to be integrated into a causal group.

        * `wipe`: Boolean to control, whether wipe(signal) should be called
         before the new spike is created.
        """
        if wipe:
            self.wipe(signal)
        with self._lock:
            self._spikes[
                Spike(sig=signal.name, parents=parents, consumable_resources=set(self._properties.keys()))] = True

    def wipe(self, signal: Signal):
        """
        Delete all spikes for the given signal. Partially fulfilled states
         that have acquired an affected spike will be forced to reject it.
        Wiping a parent spike will also wipe all child spikes.

        * `signal`: The signal for which all existing spikes (and their children)
         should be invalidated and forgotten.
        """
        with self._lock:
            for spike in self._spikes:
                if spike.name() == signal.name:
                    spike.wipe()
                    self._spikes[spike] = False
                    for child_spike in spike.offspring():
                        self._spikes[child_spike] = False
        # Final cleanup will be performed while update is running,
        #  and cg.stale(spike) returns true.
        # TODO: Make sure, that it is not a problem if the spike is currently referenced
        #  in a running state that would give it new offspring (and a second life).

    def run(self) -> None:
        """
        Creates a signal processing thread, starts it, and emits the :startup signal.
        """
        if self._run_task:
            logger.error("Attempt to start context twice!")
            return
        self._run_task = Thread(target=self._run_private)
        self._run_task.start()
        self.emit(s(":startup"))

    def shutting_down(self) -> bool:
        """
        Retrieve the shutdown flag value, which indicates whether shutdown() has been called.
        """
        return self._shutdown_flag.is_set()

    def shutdown(self) -> None:
        """
        Sets the shutdown flag and waits for the signal processing thread to join.
        """
        self._shutdown_flag.set()
        self.emit(s(":shutdown"))
        self._run_task.join()

    def add_module(self, module_name: str) -> None:
        """
        Add a module by python module folder name, or by ravestate module name.

        * `module_name`: The name of the module to be added. If it is the
         name of a python module that has not been imported yet, the python module
         will be imported, and any ravestate modules registered during the python
         import will also be added to this context.
        """
        if registry.has_module(module_name):
            self._module_registration_callback(registry.get_module(module_name))
            return
        registry.import_module(module_name=module_name, callback=self._module_registration_callback)

    def add_state(self, *, st: State) -> None:
        """
        Add a state to this context. It will be indexed wrt/ the properties/signals
         it depends on. Error messages will be generated for unknown signals/properties.

        * `st`: The state which should be added to this context.
        """
        if st in self._activations_per_state:
            logger.error(f"Attempt to add state `{st.name}` twice!")
            return

        # make sure that all of the state's depended-upon properties exist
        for prop in st.read_props+st.write_props:
            if prop not in self._properties:
                logger.error(f"Attempt to add state which depends on unknown property `{prop}`!")
                return

        # register the state's signal
        with self._lock:
            if st.signal():
                self._add_sig(st.signal())

            # add state's constraints as causes for the written prop's :changed signals,
            #  as well as the state's own signal.
            states_to_recomplete: Set[State] = {st}
            for conj in st.constraint.conjunctions():
                for propname in st.write_props:
                    if propname in self._properties:
                        for signal in self._properties[propname].signals():
                            self._signal_causes[signal].append(conj)
                            # Since a new cause for the property's signal is added,
                            #  it must be added to all states depending on that signal.
                            states_to_recomplete.update(self._states_for_signal(signal))
                if st.signal():
                    self._signal_causes[st.signal()].append(conj)

            # add state to state activation map
            self._activations_per_state[st] = set()

            # make sure that all of the state's depended-upon signals exist,
            #  add a default state activation for every constraint.
            for state in states_to_recomplete:
                self._del_state_activations(state)
                self._complete_constraint(state)
                self._new_state_activation(state)

        # register the state's consumable dummy, so that it is passed
        #  to Spike and from there to CausalGroup as a consumable resource.
        self.add_prop(prop=st.consumable)

    def rm_state(self, *, st: State) -> None:
        """
        Remove a state from this context. Note, that any state which is constrained
         on the signal that is emitted by the deleted state will also be deleted.

        * `st`: The state to remove. An error message will be generated,
         if the state was not previously added to this context with add_state().
        """
        if st not in self._activations_per_state:
            logger.error(f"Attempt to remove unknown state `{st.name}`!")
            return
        with self._lock:
            # Remove the state's signal
            if st.signal():
                self._rm_sig(st.signal())
            # Remove state activations for the state
            self._del_state_activations(st)
            # Actually forget about the state
            del self._activations_per_state[st]
        # unregister the state's consumable dummy
        self.rm_prop(prop=st.consumable)

    def add_prop(self, *, prop: PropertyBase) -> None:
        """
        Add a property to this context. An error message will be generated, if a property with
         the same name has already been added previously.

        * `prop`: The property object that should be added.
        """
        if prop.fullname() in self._properties:
            logger.error(f"Attempt to add property {prop.fullname()} twice!")
            return
        with self._lock:
            # register property
            self._properties[prop.fullname()] = prop
            # register all of the property's signals
            for signal in prop.signals():
                self._add_sig(signal)

    def rm_prop(self, *, prop: PropertyBase) -> None:
        """
        Remove a property from this context.
        Generates error message, if the property was not added with add_prop() to the context previously

        * `prop`: The property to remove.object
        """
        if prop.fullname() not in self._properties:
            logger.error(f"Attempt to remove unknown property {prop.fullname()}!")
            return
        # remove property from context
        self._properties.pop(prop.fullname())
        states_to_remove: Set[State] = set()
        with self._lock:
            # remove all of the property's signals
            for signal in prop.signals():
                self._rm_sig(signal)
            # remove all states that depend upon property
            for st in self._activations_per_state:
                if prop.fullname() in st.read_props + st.write_props:
                    states_to_remove.add(st)
        for st in states_to_remove:
            self.rm_state(st=st)

    def __getitem__(self, key: str) -> Optional[PropertyBase]:
        """
        Retrieve a property object by name, that was previously added through add_prop()
         by it's full name. The full name is always the combination of the property's
         name and it's parent's name, joined with a colon: For example, if the name
         of a property is `foo` and it belongs to the module `bar` it's full name
         will be `bar:foo`.
        An error message will be generated if no property with the given name was
         added to the context, and None will be returned.

        * `key`: The full name of the property.

        **Returns:** The property object, or None, if no property with the given name
         was added to the context.
        """
        if key not in self._properties:
            logger.error(f"Attempt to retrieve unknown property by key `{key}`!")
            return None
        return self._properties[key]

    def conf(self, *, mod: str, key: Optional[str]=None) -> Any:
        """
        Get a single config value, or all config values for a particular module.

        * `mod`: The module whose configuration should be retrieved.

        * `key`: A specific config key of the given module, if only a single
         config value should be retrieved.

        **Returns:** The value of a single config entry if key and module are both
         specified and valid, or a dictionary of config entries if only the
         module name is specified (and valid).
        """
        if key:
            return self._config.get(mod, key)
        return self._config.get_conf(mod)

    def lowest_upper_bound_eta(self, signals: Set[Signal]) -> int:
        """
        Called by activation when it is pressured to resign. The activation wants
         to know the earliest ETA of one of it's remaining required constraints.

        * `signals`: The signals, whose ETA will be calculated, and among the
         results the minimum ETA will be returned.

        **Returns:** Lowest upper bound number of ticks it should take for at least one of the required
         signals to arrive. Fixed value (1) for now.
        """
        # TODO: Proper implementation w/ state runtime_upper_bound
        return 3

    def signal_specificity(self, sig: Signal) -> float:
        """
        Called by state activation to determine it's constraint's specificity.

        * `sig`: The signal whose specificity should be returned.

        **Returns:** The given signal's specificity.
        """
        # Count activations which are interested in the signal
        if sig not in self._act_per_state_per_signal_age:
            return .0
        num_suitors = sum(
            len(acts_per_state)
            for acts_per_state in self._act_per_state_per_signal_age[sig].values())
        if num_suitors > 0:
            return 1./num_suitors
        else:
            return .0

    def reacquire(self, act: IActivation, sig: Signal):
        """
        Called by activation, to indicate, that it needs a new Spike
         for the specified signal, and should for this purpose be referenced by context.
        Note: Not thread-safe, sync must be guaranteed by caller.

        * `act`: The activation that needs a new spike of the specified nature.

        * `sig`: Signal type for which a new spike is needed.
        """
        assert isinstance(act, Activation)  # No way around it to avoid import loop
        if sig not in self._act_per_state_per_signal_age:
            logger.error(f"Attempt to reacquire for unknown signal {sig.name}!")
            return
        interested_acts = self._act_per_state_per_signal_age[sig][0][act.state_to_activate]  # sig.min_age
        interested_acts.add(act)

    def withdraw(self, act: IActivation, sig: Signal):
        """
        Called by activation to make sure that it isn't referenced
         anymore as looking for the specified signal.
        This might be, because the activation chose to eliminate itself
         due to activation pressure, or because one of the activations
         conjunctions was fulfilled, so it is no longer looking for
         signals to fulfill the remaining conjunctions.
        Note: Not thread-safe, sync must be guaranteed by caller.

        * `act`: The activation that has lost interest in the specified signal.

        * `sig`: Signal type for which interest is lost.
        """
        assert isinstance(act, Activation)  # No way around it to avoid import loop
        if sig not in self._act_per_state_per_signal_age:
            logger.error(f"Attempt to withdraw for unknown signal {sig.name}!")
            return
        interested_acts = self._act_per_state_per_signal_age[sig][0][act.state_to_activate]  # sig.min_age
        interested_acts.discard(act)

    def secs_to_ticks(self, seconds: float) -> int:
        """
        Convert seconds to an equivalent integer number of ticks,
         given this context's tick rate.

        * `seconds`: Seconds to convert to ticks.

        **Returns:** An integer tick count.
        """
        return ceil(seconds * float(self._core_config[self._tick_rate_config]))

    def _add_sig(self, sig: Signal):
        if sig in self._act_per_state_per_signal_age:
            logger.error(f"Attempt to add signal f{sig.name} twice!")
            return
        self._signal_causes[sig] = []
        self._act_per_state_per_signal_age[sig] = defaultdict(lambda: defaultdict(set))

    def _rm_sig(self, sig: Signal) -> None:
        affected_states: Set[State] = set()
        for _, acts_per_state in self._act_per_state_per_signal_age[sig].items():
            affected_states |= set(acts_per_state.keys())
        if affected_states:
            logger.warning(
                f"Since signal {sig.name} was removed, the following states will have dangling constraints: " +
                ",".join(st.name for st in affected_states))
        # Remove signal as a cause for other signals
        for _, causes in self._signal_causes.items():
            old_causes = causes.copy()
            causes.clear()
            causes += [cause for cause in old_causes if sig not in cause]
        del self._signal_causes[sig]
        self._act_per_state_per_signal_age.pop(sig)

    def _module_registration_callback(self, mod: Module):
        self._config.add_conf(mod)
        for prop in mod.props:
            self.add_prop(prop=prop)
        for st in mod.states:
            self.add_state(st=st)
        logger.info(f"Module {mod.name} added to session.")

    def _new_state_activation(self, st: State) -> None:
        activation = Activation(st, self)
        self._activations_per_state[st].add(activation)
        for signal in st.constraint_.signals():
            if signal in self._act_per_state_per_signal_age:
                # TODO: (Here and below) Determine, whether indexing by min age is actually necessary
                #  -> may not, because immediate acquisition is necessary -> otherwise, the
                #   signal's causal group will not know about a possibly higher-
                #   specificity state that runs a bit later.
                # self._act_per_state_per_signal_age[signal][signal.min_age][st] |= {activation}
                self._act_per_state_per_signal_age[signal][0][st] |= {activation}
            else:
                logger.error(
                    f"Adding state activation for f{st.name} which depends on unknown signal `{signal}`!")

    def _del_state_activations(self, st: State) -> None:
        # delete activation from gaybar
        if st not in self._activations_per_state:
            return
        for signal in st.constraint_.signals():
            if signal in self._act_per_state_per_signal_age:
                if st in self._act_per_state_per_signal_age[signal][0]:
                    del self._act_per_state_per_signal_age[signal][0][st]  # signal.min_age
        for act in self._activations_per_state[st].copy():
            act.dereference(spike=None, reacquire=False, reject=True)
            self._activations_per_state[st].remove(act)

    def _state_activations(self, *, st: Optional[State]=None) -> Set[Activation]:
        if st:
            return self._activations_per_state[st].copy()
        else:
            return {act for acts in self._activations_per_state.values() for act in acts}

    def _states_for_signal(self, sig: Signal) -> Iterable[State]:
        if sig not in self._act_per_state_per_signal_age:
            return set()
        return self._act_per_state_per_signal_age[sig][0].keys()  # sig.min_age

    def _complete_constraint(self, st: State):
        new_conjuncts: Set[Conjunct] = set()
        for conj in st.constraint.conjunctions():
            known_signals = set()
            new_conjuncts.update(
                Conjunct(*conj_signals)
                for conj_signals in self._complete_conjunction(conj, known_signals))
            assert len(known_signals) == 0
        st.constraint_ = Disjunct(*{conj for conj in new_conjuncts})

    def _complete_conjunction(self, conj: Conjunct, known_signals: Set[Signal]) -> List[Set[Signal]]:
        result = [set(conj.signals())]
        for conj_sig in conj.signals():
            completion = self._complete_signal(conj_sig, known_signals)
            if completion is not None and len(completion) > 0:
                # the signal is non-cyclic, and has at least one cause (secondary signal).
                #  permute existing disjunct conjunctions with new conjunction(s)
                result = [deepcopy(result_conj) | deepcopy(completion_conj) for result_conj in result for completion_conj in completion]
        return result

    def _complete_signal(self, sig: Signal, known_signals: Set[Signal]) -> Optional[List[Set[Signal]]]:
        # detect and handle cyclic causal chain
        if sig in known_signals:
            return None
        assert sig in self._signal_causes

        # a signal without cause (a primary signal) needs no further completion
        if not self._signal_causes[sig]:
            return []

        # a signal with at least one secondary cause needs at least one non-cyclic
        #  cause to be considered a completion itself
        result = []
        known_signals.add(sig)
        for conj in self._signal_causes[sig]:
            completion = self._complete_conjunction(conj, known_signals)
            if completion:
                result += [conj | {sig} for conj in completion]
        known_signals.discard(sig)
        return result if len(result) else None

    def _run_private(self):

        tick_interval = 1. / self._core_config[self._tick_rate_config]
        while not self._shutdown_flag.wait(tick_interval):

            with self._lock:
                # For every state, compress it's activations
                for st, acts in self._activations_per_state.items():
                    assert len(acts) > 0
                    if len(acts) > 1:
                        # We could do some fancy merge of partially fulfilled activations,
                        #  but for now let's just remove all completely unfulfilled
                        #  ones apart from one.
                        allowed_unfulfilled: Activation = None
                        for act in acts.copy():
                            if not act.spiky():
                                if allowed_unfulfilled:
                                    for signal in act.constraint.signals():
                                        if signal in self._act_per_state_per_signal_age and \
                                           act in self._act_per_state_per_signal_age[signal][0][act.state_to_activate]:
                                            self._act_per_state_per_signal_age[
                                                signal][0][act.state_to_activate].remove(act)
                                    acts.remove(act)
                                else:
                                    allowed_unfulfilled = act

                # Acquire new state activations for every spike
                for spike, acquisition_allowed in self._spikes.items():
                    if not acquisition_allowed:
                        continue
                    for state, acts in self._act_per_state_per_signal_age[s(spike.name())][spike.age()].items():
                        old_acts = acts.copy()
                        for act in old_acts:
                            if act.acquire(spike):
                                # Remove the Activation instance from _act_per_state_per_signal_age
                                #  for the Spike with a certain minimum age. In place of the removed
                                #  activation, if no activation with the same target state is left,
                                #  a new Activation will be created.
                                acts.remove(act)
                                if len(acts) == 0:
                                    self._new_state_activation(state)
                            else:
                                logger.error(
                                    "An activation rejected a spike it was registered to be interested in.")

                # Update all state activations.
                for act in self._state_activations():
                    if act.update():
                        self._activations_per_state[act.state_to_activate].discard(act)

                # Forget fully unreferenced spikes
                old_spike_set = self._spikes.copy()
                for spike in old_spike_set:
                    with spike.causal_group() as cg:
                        if cg.stale(spike):
                            # This should lead to the deletion of the spike
                            self._spikes.pop(spike)
                            spike.wipe(already_wiped_in_causal_group=True)
                            logger.debug(f"{cg}.stale({spike.name()})->Y")

                # Increment age on active spikes
                for spike in self._spikes:
                    spike.tick()

                # Force garbage collect
                gc.collect()



from ravestate import registry
from ravestate_ros2.ros2_properties import sync_ros_properties

registry.register(
    name="ros2",
    states=(sync_ros_properties,),
    config={
        # name of the ROS2-Node that is created by ravestate_ros2
        ros2_properties.NODE_NAME_CONFIG_KEY: "ravestate_ros2",
        # frequency for spinning of ROS2-Node in spins per second (0: as fast as possible)
        ros2_properties.SPIN_FREQUENCY_CONFIG_KEY: 10
    })



import time
from typing import Dict

from ravestate.state import state, Delete
from ravestate.constraint import s
from ravestate.property import PropertyBase

from ravestate.receptor import receptor
from ravestate.wrappers import ContextWrapper

from reggol import get_logger
logger = get_logger(__name__)

ROS2_AVAILABLE = True
try:
    import rclpy
except ImportError:
    logger.error("Could not import rclpy. Please make sure to have ROS2 installed.")
    ROS2_AVAILABLE = False

NODE_NAME_CONFIG_KEY = "ros2-node-name"
SPIN_FREQUENCY_CONFIG_KEY = "ros2-spin-frequency"

global_prop_set = set()


@state(cond=s(":startup"))
def sync_ros_properties(ctx: ContextWrapper):
    """
    State that creates a ROS2-Node, registers all Ros2SubProperties and Ros2PubProperties in ROS2 and keeps them synced
    """
    # check for ROS2 availability
    if not ROS2_AVAILABLE:
        logger.error("ROS2 is not available, therefore all ROS2-Properties "
                     "will be just normal properties without connection to ROS2!")
        return Delete()

    # get config stuff
    node_name = ctx.conf(key=NODE_NAME_CONFIG_KEY)
    if not node_name:
        logger.error(f"{NODE_NAME_CONFIG_KEY} is not set. Shutting down ravestate_ros2")
        return Delete()
    spin_frequency = ctx.conf(key=SPIN_FREQUENCY_CONFIG_KEY)
    if spin_frequency is None or spin_frequency < 0:
        logger.error(f"{SPIN_FREQUENCY_CONFIG_KEY} is not set or less than 0. Shutting down ravestate_ros2")
        return Delete()
    if spin_frequency == 0:
        spin_sleep_time = 0
    else:
        spin_sleep_time = 1 / spin_frequency

    # init ROS
    rclpy.init()
    node = rclpy.create_node(node_name)

    global global_prop_set
    # current_props: hash -> subscription/publisher
    current_props: Dict = dict()

    # ROS-Context Sync Loop
    while not ctx.shutting_down():
        # remove deleted props
        removed_props = current_props.keys() - global_prop_set
        for prop_hash in removed_props:
            item = current_props[prop_hash]
            if isinstance(item, rclpy.subscription.Subscription):
                node.destroy_subscription(item)
            elif isinstance(item, rclpy.publisher.Publisher):
                node.destroy_publisher(item)
            elif isinstance(item, rclpy.client.Client):
                node.destroy_client(item)
            current_props.pop(prop_hash)

        # add new props
        new_props = global_prop_set - current_props.keys()
        for prop in new_props:
            # register subscribers in ROS
            if isinstance(prop, Ros2SubProperty):
                # register in context
                @receptor(ctx_wrap=ctx, write=prop.fullname())
                def ros_to_ctx_callback(ctx, msg, prop_name: str):
                    ctx[prop_name] = msg

                prop.ros_to_ctx_callback = ros_to_ctx_callback
                prop.subscription = node.create_subscription(prop.msg_type, prop.topic, prop.ros_subscription_callback)
                current_props[prop.__hash__()] = prop.subscription
            # register publishers in ROS
            if isinstance(prop, Ros2PubProperty):
                prop.publisher = node.create_publisher(prop.msg_type, prop.topic)
                current_props[prop.__hash__()] = prop.publisher
            # register clients in ROS
            if isinstance(prop, Ros2CallProperty):
                prop.client = node.create_client(prop.service_type, prop.service_name)
                current_props[prop.__hash__()] = prop.client

            # replace prop with hash in global_props
            global_prop_set.remove(prop)
            global_prop_set.add(prop.__hash__())

        # spin once
        rclpy.spin_once(node, timeout_sec=0)
        time.sleep(spin_sleep_time)

    node.destroy_node()
    rclpy.shutdown()


class Ros2SubProperty(PropertyBase):
    def __init__(self, name: str, topic: str, msg_type, default_value=None, always_signal_changed: bool = True):
        """
        Initialize Property

        * `name`: Name of the property

        * `topic`: ROS2-Topic that should be subscribed

        * `msg_type`: ROS2-Message type of messages in the topic

        * `default_value`: Default value of the property

        * `always_signal_changed`: if signal:changed should be emitted if value is written again without changing
        """
        super().__init__(
            name=name,
            allow_read=True,
            allow_write=True,
            allow_push=False,
            allow_pop=False,
            default_value=default_value,
            always_signal_changed=always_signal_changed)
        self.topic = topic
        self.msg_type = msg_type
        self.subscription = None
        self.ros_to_ctx_callback = None
        global global_prop_set
        global_prop_set.add(self)

    def __del__(self):
        global global_prop_set
        global_prop_set.remove(self.__hash__())

    def ros_subscription_callback(self, msg):
        """
        Writes the message from ROS to the property

        * `msg`: ROS2-Message that should be written into the property
        """
        if self.ros_to_ctx_callback:
            logger.debug(f"{self.fullname()} received message {str(msg)} from topic {self.topic}")
            self.ros_to_ctx_callback(msg=msg, prop_name=self.fullname())


class Ros2PubProperty(PropertyBase):
    def __init__(self, name: str, topic: str, msg_type):
        """
        Initialize Property

        * `name`: Name of the property

        * `topic`: ROS2-Topic that messages should be pubished to

        * `msg_type`: ROS2-Message type of messages in the topic
        """
        super().__init__(
            name=name,
            allow_read=True,
            allow_write=True,
            allow_push=False,
            allow_pop=False,
            default_value=None,
            always_signal_changed=False)
        self.topic = topic
        self.msg_type = msg_type
        self.publisher = None
        global global_prop_set
        global_prop_set.add(self)

    def __del__(self):
        global global_prop_set
        global_prop_set.remove(self.__hash__())

    def write(self, value):
        """
        Publish value on ROS

        * `value`: ROS2-Message that should be published
        """
        if super().write(value):
            if self.publisher:
                logger.debug(f"{self.fullname()} is publishing message {str(value)} on topic {self.topic}")
                self.publisher.publish(value)
            elif ROS2_AVAILABLE:
                logger.error(f"Message {str(value)} on topic {self.topic} "
                             f"cannot be published because publisher was not registered in ROS")


class Ros2CallProperty(PropertyBase):
    def __init__(self, name: str, service_name: str, service_type, call_timeout: float = 10.0):
        """
        Initialize Property

        * `name`: Name of the property

        * `service_name`: ROS2-Service that should be called

        * `service_type`: Type of the service used

        * `call_timeout`: Timeout when waiting for availability of a service in seconds
        """
        super().__init__(
            name=name,
            allow_read=True,
            allow_write=True,
            allow_push=False,
            allow_pop=False,
            default_value=None,
            always_signal_changed=False)
        self.service_name = service_name
        self.service_type = service_type
        self.call_timeout = call_timeout
        self.client = None
        global global_prop_set
        global_prop_set.add(self)

    def __del__(self):
        global global_prop_set
        global_prop_set.remove(self.__hash__())

    def write(self, value):
        """
        Call Service and receive result directly in the property.
        Blocks during service-call.
        If service is not available, writes None into the property value and returns.

        * `value`: Either Request-Object for the service or dict containing the attributes of the Request
        """
        if super().write(value):
            if self.client:
                if not self.client.wait_for_service(timeout_sec=self.call_timeout):
                    logger.error(f'service {self.service_name} not available')
                    super().write(None)
                    return

                if isinstance(value, self.service_type.Request):
                    # value is already Request
                    req = value
                else:
                    # value is dict {"param1": value1, "param2": value2}
                    req = self.service_type.Request(**value)

                logger.debug(f"{self.fullname()} is sending request {str(req)} to service {self.service_name}")
                result = self.client.call(req)
                super().write(result)

            elif ROS2_AVAILABLE:
                logger.error(f"Request {str(value)} to service {self.service_name} "
                             f"cannot be sent because client was not registered in ROS")

